# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import


"""
Warning defined here are for conditions that can affect the results of a calculation,
but do not prevent a calculation from proceeding without error.

API warnings about the use of default values in calculations are handled programmatically
within each class.
"""

# these are warning messages that are common to all modules. the specific warnings currently
# defined will probably go away once API and sanity checking is fully implemented and verified
# throughout the system. these messages get combined with each of the module-specific messages using
# the update() method.
standard_warning_messages = {
    "no_sanity_check": "No sanity checking implemented for %s: %s",
    "no_api_check": "No API checking implemented for %s: %s"
}

# warning messages specific to etc3D and its classes
etc3d_warning_messages = {
    "background_saturated": "Background is saturating. Decrease Groups and try increasing Integrations or Exposures."
}

# warning messages specific to Strategy and its sub-classes.
strategy_warning_messages = {
    "background_region_too_small": "Background region smaller than source extraction region. This can adversely affect the SNR.",
    "extraction_aperture_truncated": "Extraction aperture partially outside of the field of view.",
    "background_region_truncated": "Background estimation region partially outside of the field of view.",
    "extraction_aperture_undersampled": "Extraction aperture undersampled. Pixel area %.2f%% less than requested area.",
    "background_region_undersampled": "Background estimation region undersampled. Pixel area %.2f%% less than requested area.",
    "upper_background_region_missing": "Upper background estimation region outside of field of view.",
    "upper_background_region_truncated": "Upper background estimation region truncated by %.2f%%.",
    "lower_background_region_missing": "Lower background estimation region outside of field of view.",
    "lower_background_region_truncated": "Lower background estimation region truncated by %.2f%%.",
    "target_x_position_not_used": "Target X position is not meaningful and thus not used in spectral aperture photometry.",
    "target_occulted": "Specified target position is occulted by the coronagraphy mask.",
    "coronagraphy_central_source": "Coronagraphy requires a central source within {0:.2f} of the center.",
    "coronagraphy_psf_source": "Coronagraphy requires the PSF subtraction source to be outside the scene (>{0:.2f} arcsec, not {1:.2f}"
}
strategy_warning_messages.update(standard_warning_messages)

# warning messages specific to Normalization and its sub-classes
normalization_warning_messages = {
    "normalized_to_zero_flux": "Zero flux at reference wavelength. Spectrum left unscaled.",
    "unsupported_normalization_bandpass": "Bandpass specification, %s, not currently supported, but may work."
}
normalization_warning_messages.update(standard_warning_messages)

# warning messages specific to SED and its sub-classes. currently none defined...
sed_warning_messages = {}
sed_warning_messages.update(standard_warning_messages)

# warning messages specific to Telescope and its sub-classes
telescope_warning_messages = {
    "telescope_ote_efficiency_missing": "Telescope OTE throughput mis-configured or unavailable. Using default value of 1.0.",
    "telescope_background_missing": "Telescope notional background mis-configured or unavailable. Using default value of 0.0."
}
telescope_warning_messages.update(standard_warning_messages)

# warning messages specific to AstroSpectrum and its classes and methods
astrospectrum_warning_messages = {
    "max_scene_size_reached": "Scene requires a total field-of-view of %.3f arcsec. Using the configured maximum of %.3f arcsec.",
    "scene_fov_too_small": "Field-of-view size, %.3f, too small to encompass any of the defined sources.",
    "wavelength_truncated_blue": "Spectrum blue limit, %.2f, does not extend to instrument configuration blue limit, %.2f.",
    "wavelength_truncated_red": "Spectrum red limit, %.2f, does not extend to instrument configuration red limit, %.2f.",
    "scene_range_truncated": "Combined wavelength range of scene [%.2f, %.2f] less than instrument configuration's range [%.2f, %.2f].",
    "spectrum_missing_blue": "Spectrum [%.2f, %.2f] does not extend to instrument configuration blue limit, %.2f.",
    "spectrum_missing_red": "Spectrum [%.2f, %.2f] does not extend to instrument configuration red limit, %.2f."
}
astrospectrum_warning_messages.update(standard_warning_messages)

report_warning_messages = {
    "partial_saturated": "Partial saturation:\n There are {0:1d} pixels saturated at the end of a ramp. Partial ramps may still be used in some cases.",
    "full_saturated": "Full saturation:\n There are {0:1d} pixels saturated at the end of the first group. These pixels cannot be recovered.",
    "ifu_partial_saturated": "Partial saturation:\n In the current wavelength slice, there are {0:1d} pixels saturated at the end of a ramp. Partial ramps may still be used in some cases.",
    "ifu_full_saturated": "Full saturation:\n In the current wavelength slice, there are {0:1d} pixels saturated at the end of the first group. These pixels cannot be recovered.",
    "cube_partial": "There are {0:} total partially saturated pixels in the data cube.",
    "cube_full": "There are {0:} total fully saturated pixels in the data cube.",
    "bad_waveref": "Specified wavelength, {0:.2f}, out of range [{1:.2f}, {2:.2f}]. Using {3:.2f} to select diagnostic planes instead.",
    "ta_max_saturated_pixels": "<font color=red><b>TA MAY FAIL</b></font>: Number of fully saturated pixels {0:} in centroid" \
                    " box exceeds the maximum number of {1:} pixels allowed to ensure a successful target acquisition. "\
                    "Recommend to adjust Detector Setup and/or Instrument Setup accordingly.",
    "ta_snr_threshold": "<font color=red><b>TA MAY FAIL</b></font>: calculated SNR {0:.2f} must exceed {1:.2f} to ensure TA success.",
    "partial_saturated_ami": "Partial saturation:\n There are {0:1d} pixels saturated* at the end of a ramp. Partial ramps may still be used in some cases. *(See footnote in the Saturation image tab.)",
    "full_saturated_ami": "Full saturation:\n There are {0:1d} pixels saturated* at the end of the first group. These pixels cannot be recovered. *(See footnote in the Saturation image tab.)"
}
report_warning_messages.update(standard_warning_messages)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import copy
import numpy as np
from warnings import warn
import scipy.integrate as integrate
import scipy.interpolate as sci_int
from scipy.ndimage.interpolation import shift
import scipy.signal as sg
from astropy.io import fits

from . import observation
from . import astro_spectrum as astro
from . import background as bg
from . import coords
from .config import DefaultConfig
from .report import ReportFactory
from .scene import Scene
from .calc_utils import build_empty_scene
from .custom_exceptions import EngineInputError, EngineOutputError, RangeError, DataError
from .instrument_factory import InstrumentFactory
from .strategy import StrategyFactory
from .pandeia_warnings import etc3d_warning_messages as warning_messages
from . import debug_utils

from six.moves import zip


# pyfftw.interfaces.cache.enable()

class CalculationConfig(DefaultConfig):
    """
    Encapsulate calculation configuration parameters (e.g. effects to include, noise sources to consider)
    """
    pass


class DetectorSignal(astro.ConvolvedSceneCube):
    """
    This class contains functionality for calculating the integrated electron rate of an
    astronomical source at the relevant instrument detector plane.

    Parameters
    ----------
    observation: observation.Observation instance
        Contains information required to configure a calculation
    calc_config: CalculationConfig instance
        Contains boolean flags that control which noise components are included in the calculation
    webapp: bool
        Toggle strict engine API checking
    order: int or None
        For multi-order case (i.e. SOSS), designate which order to use

    Attributes
    ----------
    """

    def __init__(self, observation, calc_config=CalculationConfig(), webapp=False, order=None, empty_scene=False):
        # Get calculation configuration
        self.calculation_config = calc_config

        # Link to the passed observation
        self.observation = observation

        # Load the instrument we're using
        self.current_instrument = observation.instrument
        # and configure it for the order we wish to use, if applicable
        self.current_instrument.order = order
        # save to the DetectorSignal instance, for convenience purposes
        self.order = order

        # how are we projecting the signal onto the detector plane?
        self.projection_type = self.current_instrument.projection_type

        # If we're in a dispersed mode, we need to know which axis the signal is dispersed along
        self.dispersion_axis = self.current_instrument.dispersion_axis()

        # Get the detector parameters (read noise, etc.)
        self.det_pars = self.current_instrument.get_detector_pars()

        # Initialize detector mask
        self.det_mask = 1.0

        # Get the background
        self.background = bg.Background(self.observation, webapp=webapp)

        # Then initialize the flux and wavelength grid
        astro.ConvolvedSceneCube.__init__(
            self,
            self.observation.scene,
            self.current_instrument,
            background=self.background,
            psf_library=self.current_instrument.psf_library,
            webapp=webapp,
            empty_scene=empty_scene
        )

        self.warnings.update(self.background.warnings)
        # We have to propagate the background through the system transmission
        # to get the background in e-/s/pixel/micron. The background rate is a 1D spectrum.
        self.bg_fp_rate = self.get_bg_fp_rate()

        # Initialize slice lists
        self.rate_list = []
        self.rate_plus_bg_list = []
        self.saturation_list = []
        self.groups_list = []
        self.pixgrid_list = []

        # Loop over all slices and calculate the photon and electron rates through the
        # observatory for each one. Note that many modes (imaging, etc.) will have just
        # a single slice.
        for flux_cube, flux_plus_bg in zip(self.flux_cube_list, self.flux_plus_bg_list):
            # Rates for the slice without the background
            slice_rate = self.all_rates(flux_cube, add_extended_background=False)

            # Rates for the slice with the background added
            slice_rate_plus_bg = self.all_rates(flux_plus_bg, add_extended_background=True)

            # Saturation map for the slice
            slice_saturation = self.get_saturation_mask(rate=slice_rate_plus_bg['fp_pix'])
            slice_group = self.current_instrument.exposure_spec.get_groups_before_sat(slice_rate_plus_bg['fp_pix'],
                                                                                      self.det_pars['fullwell'])

            # The grid in the slice
            slice_pixgrid = self.get_pix_grid(slice_rate)

            # Append all slices to the master lists
            self.rate_list.append(slice_rate)
            self.rate_plus_bg_list.append(slice_rate_plus_bg)
            self.saturation_list.append(slice_saturation)
            self.groups_list.append(slice_group)
            self.pixgrid_list.append(slice_pixgrid)

        # Get the mapping of wavelength to pixels on the detector plane. This is grabbed from the
        # first entry in self.rate_list and is currently defined to be the same for all slices.
        self.wave_pix = self.get_wave_pix()

        # This is also grabbed from the first slice as a diagnostic
        self.fp_rate = self.get_fp_rate()

        # Note that the 2D image due to background alone may have spatial structure due to instrumental effects.
        # Therefore it is calculated here.
        self.bg_pix_rate = self.get_bg_pix_rate()

        # Check to see if the background is saturating
        bgsat = self.get_saturation_mask(rate=self.bg_pix_rate)
        if (np.sum(bgsat) > 0) or (np.isnan(np.sum(bgsat))):
            key = "background_saturated"
            self.warnings[key] = warning_messages[key]

        # Reassemble rates of multiple slices on the detector
        self.rate = self.on_detector(self.rate_list)
        self.rate_plus_bg = self.on_detector(self.rate_plus_bg_list)
        self.ngroup_map = self.current_instrument.exposure_spec.get_groups_before_sat(self.rate_plus_bg,
                                                                                      self.det_pars['fullwell'])
        self.fraction_saturation = np.max(
            self.current_instrument.exposure_spec.get_saturation_fraction(self.rate_plus_bg,
                                                                          self.det_pars['fullwell']))
        self.detector_pixels = self.current_instrument.get_detector_pixels(self.wave_pix)
        self.brightest_pixel = np.max(self.rate_plus_bg)

        # Get the read noise correlation matrix and store it as an attribute.
        if self.det_pars['rn_correlation']:
            self.read_noise_correlation_matrix = self.current_instrument.get_readnoise_correlation_matrix(
                self.rate.shape)

    def spectral_detector_transform(self):
        """
        Create engine API format dict section containing properties of wavelength coordinates
        at the detector plane.

        Returns
        -------
        t: dict (engine API compliant keys)
        """
        t = {}
        t['wave_det_refpix'] = 0
        t['wave_det_max'] = self.wave_pix.max()
        t['wave_det_min'] = self.wave_pix.min()

        # there are currently three projection_type's which are basically detector plane types:
        #
        # 'spec' - where the detector plane is purely dispersion vs. spatial
        # 'slitless' - basically a special case of 'spec' with where dispersion and spatial are mixed
        # 'image' - where the detector plane is purely spatial vs. spatial (i.e. no disperser element)
        #
        # 'IFU' mode is of projection_type='spec' because the mapping from detector X pixels to
        # wavelength is the same for each slice.  this projection_type will work for 'MSA' mode as well
        # because we will only handle one aperture at a time.  'slitless' spectroscopy will mix
        # spatial and dispersion information onto the detector X axis.  however, the detector
        # plane is fundamentally spatial vs. wavelength in that case so it's handled the same as
        # projection_type='spec'. creating a spectrum for a specific target will be handled via the
        # extraction strategy.
        if self.projection_type in ('spec', 'slitless', 'multiorder'):
            t['wave_det_size'] = len(self.wave_pix)
            if len(self.wave_pix) > 1:
                # we don't yet have a way of handling non-linear coordinate transforms here. that said,
                # this is mostly right for most of our cases with nirspec prism being the notable exception.
                # this is also only used for plotting purposes while the true actual wave_pix mapping is used
                # internally for all calculations.
                t['wave_det_step'] = (self.wave_pix[-1] - self.wave_pix[0]) / t['wave_det_size']
            else:
                t['wave_det_step'] = 0.0
            t['wave_det_refval'] = self.wave_pix[0]
        elif self.projection_type == "image":
            t['wave_det_step'] = 0.0
            t['wave_det_refval'] = self.wave_pix[0]
            t['wave_det_size'] = 1
        else:
            message = "Unsupported projection_type: %s" % self.projection_type
            raise EngineOutputError(value=message)
        return t

    def wcs_info(self):
        """
        Get detector coordinate transform as a dict of WCS keyword/value pairs.

        Returns
        -------
        header: dict
            WCS header keys defining coordinate transform in the detector plane
        """
        if self.projection_type == 'image':
            # if we're in imaging mode, the detector sampling is the same as the model
            header = self.grid.wcs_info()
        elif self.projection_type in ('spec', 'slitless', 'multiorder'):
            # if we're in a dispersed mode, dispersion can be either along the X or Y axis. the image outputs in
            # the engine Report are rotated so that dispersion will always appear to be along the X axis with
            # wavelength increasing with increasing X (i. e. dispersion angle of 0).  currently, the only other
            # supported dispersion angle is 90 which is what we get when dispersion_axis == 'y'.
            t = self.grid.as_dict()
            t.update(self.spectral_detector_transform())
            header = {
                'ctype1': 'Wavelength',
                'crpix1': 1,
                'crval1': t['wave_det_min'] - 0.5 * t['wave_det_step'],
                'cdelt1': t['wave_det_step'],
                'cunit1': 'um',
                'cname1': 'Wavelength',
                'ctype2': 'Y offset',
                'crpix2': 1,
                'crval2': t['y_min'] - 0.5 * t['y_step'],
                'cdelt2': -t['y_step'],
                'cunit2': 'arcsec',
                'cname2': 'Detector Offset',
            }
            if self.dispersion_axis == 'y':
                header['ctype2'] = 'X offset'
                header['crval2'] = t['x_min'] - 0.5 * t['x_step'],
                header['cdelt2'] = t['x_step']
        else:
            message = "Unsupported projection_type: %s" % self.projection_type
            raise EngineOutputError(value=message)
        return header

    def get_wave_pix(self):
        """
        Return the mapping of wavelengths to pixels on the detector plane
        """
        return self.rate_list[0]['wave_pix']

    def get_fp_rate(self):
        """
        Return scene flux at the focal plane in e-/s/pixel/micron (excludes background)
        """
        return self.rate_list[0]['fp']

    def get_bg_fp_rate(self):
        """
        Calculate background in e-/s/pixel/micron at the focal plane. Also correct for any excess in predicted background
        if there are pupil losses in the PSF. (#2529)
        """
        bg_fp_rate = self.focal_plane_rate(self.ote_rate(self.background.mjy_pix))
        wave_range = self.current_instrument.get_wave_range()
        pupil_thru = self.current_instrument.psf_library.get_pupil_throughput(wave_range['wmin'],
                                                                              self.current_instrument.instrument[
                                                                                  'instrument'],
                                                                              self.current_instrument.instrument[
                                                                                  'aperture'])
        return bg_fp_rate * pupil_thru

    def get_bg_pix_rate(self):
        """
        Calculate the background on the detector in e-/s/pixel
        """
        bg_pix_rate = self.rate_plus_bg_list[0]['fp_pix'] - self.rate_list[0]['fp_pix']
        return bg_pix_rate

    def on_detector(self, rate_list):
        """
        This will take the list of (pixel) rates and use them create a single detector frame. A single
        image will only have one rate in the list, but the IFUs will have n_slices. There may be other examples,
        such as different spectral orders for NIRISS. It is not yet clear how many different flavors there are, so
        this step may get refactored if it gets too complicated. Observing modes that only have one set of rates
        (imaging and single-slit spectroscopy, for instance) will still go through this, but the operation is trivial.
        """
        aperture_sh = rate_list[0]['fp_pix'].shape
        n_apertures = len(rate_list)
        detector_shape = (aperture_sh[0] * n_apertures, aperture_sh[1])
        detector = np.zeros(detector_shape)

        i = 0
        for rate in rate_list:
            detector[i * aperture_sh[0]:(i + 1) * aperture_sh[0], :] = rate['fp_pix']
            i += 1

        return detector

    def get_pix_grid(self, rate):
        """
        Generate the coordinate grid of the detector plane
        """
        if self.projection_type == 'image':
            grid = self.grid
        elif self.projection_type in ('spec', 'slitless', 'multiorder'):
            nw = rate['wave_pix'].shape[0]
            if self.dispersion_axis == 'x':
                # for slitless calculations, the dispersion axis is longer than the spectrum being dispersed
                # because the whole field of view is being dispersed. 'excess' is the size of the FOV
                # and half will be to the left of the blue end of the spectrum and half to the right of the red end.
                # this is used to create the new spatial coordinate transform for the pixel image on the detector.
                excess = rate['fp_pix'].shape[1] - nw
                pix_grid = coords.IrregularGrid(
                    self.grid.col,
                    (np.arange(nw + excess) - (nw + excess) / 2.0) * self.grid.xsamp
                )
            else:
                excess = rate['fp_pix'].shape[0] - nw
                pix_grid = coords.IrregularGrid(
                    (np.arange(nw + excess) - (nw + excess) / 2.0) * self.grid.ysamp,
                    self.grid.row
                )
            return pix_grid
        else:
            raise EngineOutputError(value="Unsupported projection_type: %s" % self.projection_type)
        return grid

    def all_rates(self, flux, add_extended_background=False):
        """
        Calculate rates in e-/s/pixel/micron or e-/s/pixel given a flux cube in mJy

        Parameters
        ----------
        flux: ConvolvedSceneCube instance
            Convolved source flux cube with flux units in mJy
        add_extended_background: bool (default=False)
            Toggle for including extended background not contained within the flux cube

        Returns
        -------
        products: dict
            Dict of products produced by rate calculation.
                'wave_pix' - Mapping of wavelength to detector pixels
                'ote' - Source rate at the telescope aperture
                'fp' - Source rate at the focal plane in e-/s/pixel/micron
                'fp_pix' - Source rate per pixel
                'fp_pix_no_ipc' - Source rate per pixel excluding effects if inter-pixel capacitance
        """
        # The source rate at the telescope aperture
        ote_rate = self.ote_rate(flux)

        # The source rate at the focal plane in interacting photons/s/pixel/micron
        fp_rate = self.focal_plane_rate(ote_rate)

        # the fp_pix_variance is the variance of the per-pixel electron rate and includes the chromatic effects
        # of quantum yield.
        if self.projection_type == 'image':
            # The wavelength-integrated rate in e-/s/pixel, relevant for imagers
            fp_pix_rate, fp_pix_variance = self.image_rate(fp_rate)
            wave_pix = self.wave_eff(fp_rate)

        elif self.projection_type == 'spec':
            # The wavelength-integrated rate in e-/s/pixel, relevant for spectroscopy
            wave_pix, fp_pix_rate, fp_pix_variance = self.spec_rate(fp_rate)

        elif self.projection_type in ('slitless', 'multiorder'):
            # The wavelength-integrated rate in e-/s/pixel, relevant for slitless spectroscopy
            wave_pix, fp_pix_rate, fp_pix_variance = self.slitless_rate(
                fp_rate,
                add_extended_background=add_extended_background
            )

        else:
            raise EngineOutputError(value="Unsupported projection_type: %s" % self.projection_type)

        # Include IPC effects, if available and requested
        if self.det_pars['ipc']:
            kernel = self.current_instrument.get_ipc_kernel()
            fp_pix_rate_ipc = self.ipc_convolve(fp_pix_rate, kernel)
        else:
            fp_pix_rate_ipc = fp_pix_rate

        # fp_pix is the final product. Since there is no reason to
        # carry around the ipc label everywhere, we rename it here.
        products = {
            'wave_pix': wave_pix,
            'ote': ote_rate,
            'fp': fp_rate,
            'fp_pix': fp_pix_rate_ipc,
            'fp_pix_no_ipc': fp_pix_rate,  # this is for calculating saturation
            'fp_pix_variance': fp_pix_variance  # this is for calculating the detector noise
        }
        return products

    def ote_rate(self, flux):
        """
        Calculate source rate in e-/s/pixel/micron at the telescope entrance aperture given
        a flux cube in mJy/pixel.
        """
        # spectrum in mJy/pixel, wave in micron, f_lambda in photons/cm^2/s/micron
        f_lambda = 1.5091905 * (flux / self.wave)
        ote_int = self.current_instrument.telescope.get_ote_eff(self.wave)
        coll_area = self.current_instrument.telescope.coll_area
        a_lambda = coll_area * ote_int
        # e-/s/pixel/micron
        ote_rate = f_lambda * a_lambda
        return ote_rate

    def focal_plane_rate(self, rate):
        """
        Takes the output from self.ote_rate() and multiplies it by the components
        of efficiency within the system and returns the source rate at the focal plane in
        e-/s/pixel/micron.
        """
        filter_eff = self.current_instrument.get_filter_eff(self.wave)
        disperser_eff = self.current_instrument.get_disperser_eff(self.wave)
        internal_eff = self.current_instrument.get_internal_eff(self.wave)
        qe = self.current_instrument.get_detector_qe(self.wave)

        fp_rate = rate * filter_eff * disperser_eff * internal_eff * qe
        return fp_rate

    def spec_rate(self, rate):
        '''
        For slitted spectrographs, calculate the detector signal by integrating
        along the dispersion direction of the cube (which is masked by a, by assumption,
        narrow slit). For slitless systems or slits wider than the PSF, the slitless_rate
        method should be used to preserve spatial information within the slit.

        Parameters
        ---------
        rate: numpy.ndarray
            Rate of photons interacting with detector as a function of model wavelength set

        Returns
        -------
        products: 3-element tuple of numpy.ndarrays
            first element - map of pixel to wavelength
            second element - electron rate per pixel
            third element - variance of electron rate per pixel
        '''
        dispersion = self.current_instrument.get_dispersion(self.wave)
        wave_pix = self.current_instrument.get_wave_pix()
        wave_pix_trunc = wave_pix[np.where(np.logical_and(wave_pix >= self.wave.min(),
                                                          wave_pix <= self.wave.max()))]

        # Check that the source spectrum is actually inside the instrumental wavelength
        # coverage.
        if len(wave_pix_trunc) == 0:
            raise RangeError(value='wave and wave_pix do not overlap')

        # Check the dispersion axis to determine which axis to sum and interpolate over
        if self.dispersion_axis == 'x':
            axis = 1
        else:
            axis = 0

        # We can simply sum over the dispersion direction. This is where we lose the spatial information within the aperture.
        spec_rate = np.sum(rate, axis=axis)

        # And then scale to the dispersion function (pixel/micron) to transform
        # from e-/s/micron to e-/s/pixel.
        spec_rate_pix = spec_rate * dispersion

        # but we are still sampled on the internal grid, so we have to interpolate to the pixel grid.
        # use kind='slinear' since it's ~2x more memory efficient than 'linear'. 'slinear' uses different code path to
        # calculate the slopes.
        int_spec_rate = sci_int.interp1d(self.wave, spec_rate_pix, axis=axis, kind='slinear', assume_sorted=True,
                                         copy=False)
        spec_rate_pix_sampled = int_spec_rate(wave_pix_trunc)

        # Handle a detector gap here by constructing a mask. If the current_instrument implements it,
        # it'll be a real mask array.  Otherwise it will simply be 1.0.
        self.det_mask = self.current_instrument.create_gap_mask(wave_pix_trunc)

        # this is the interacting photon rate in the detector with mask applied.
        spec_rate_pix_sampled *= self.det_mask

        # Add effects of non-unity quantum yields. For the spec projection, we assume that the quantum yield does not
        # change over a spectral element. Then we can just multiply the products by the relevant factors.
        q_yield, fano_factor = self.current_instrument.get_quantum_yield(wave_pix_trunc)

        # convert the photon rate to electron rate by multiplying by the quantum yield which is a function of wavelength
        spec_electron_rate_pix = spec_rate_pix_sampled * q_yield

        # to meet IDT expectations, some instruments require a possibly chromatic fudge factor to be applied
        # to the per-pixel electron rate variance.
        var_fudge = self.current_instrument.get_variance_fudge(wave_pix_trunc)

        # the variance in the electron rate, Ve, is also scaled by the quantum yield plus a fano factor which is
        # analytic in the simple 1 or 2 electron case: Ve = (qy + fano) * Re.  since Re is the photon rate
        # scaled by the quantum yield, Re = qy * Rp, we get: Ve = qy * (qy + fano) * Rp
        spec_electron_variance_pix = spec_rate_pix_sampled * q_yield * (q_yield + fano_factor) * var_fudge
        products = wave_pix_trunc, spec_electron_rate_pix, spec_electron_variance_pix

        return products

    def image_rate(self, rate):
        '''
        Calculate the electron rate for imaging modes by integrating along
        the wavelength direction of the cube.

        Parameters
        ---------
        rate: numpy.ndarray
            Rate of photons interacting with detector as a function of model wavelength set

        Returns
        -------
        products: 2-element tuple of numpy.ndarrays
            first element - electron rate per pixel
            second element - variance of electron rate per pixel
        '''
        q_yield, fano_factor = self.current_instrument.get_quantum_yield(self.wave)

        # convert the photon rate to electron rate by multiplying by the quantum yield which is a function of wavelength
        electron_rate_pix = integrate.simps(rate * q_yield, self.wave)

        # to meet IDT expectations, some instruments require a possibly chromatic fudge factor to be applied
        # to the per-pixel electron rate variance.
        var_fudge = self.current_instrument.get_variance_fudge(self.wave)

        # the variance in the electron rate, Ve, is also scaled by the quantum yield plus a fano factor which is
        # analytic in the simple 1 or 2 electron case: Ve = (qy + fano) * Re.  since Re is the photon rate
        # scaled by the quantum yield, Re = qy * Rp, we get: Ve = qy * (qy + fano) * Rp
        electron_variance_pix = integrate.simps(rate * q_yield * (q_yield + fano_factor) * var_fudge, self.wave)

        products = electron_rate_pix, electron_variance_pix

        return products

    def slitless_rate(self, rate, add_extended_background=True):
        '''
        Calculate the detector rates for slitless modes. Here we retain all spatial information and build
        up the detector plane by shifting and coadding the frames from the convolved flux cube. Also need to handle
        and add background that comes from outside the flux cube, but needs to be accounted for.

        How this works:
        1. We calculate wave_pix and truncate it to the wavelength range of the instrument (which should already 
        have been done, but it is a check)
        2. We grab the detector_pixel (dispersion) and trace (cross-dispersion) pixel locations, interpolated per
        wavelength on the detector. Except for modes with specific tracefile that map wavelength to 
        (detector_pixel,trace), these will simply be numerical and constant-0 files specifying the upper left 
        corner of where to put the wavelength grid.
        3. We then step through wavelengths. If there IS a trace file, we shift in the cross dispersion direction
        and add the rate cube plane to the image at that location. Otherwise, no expensive shifting operation is
        necessary. 
        4. If extended background is switched on, we add background for that wavelength to the rest of the detector
        image not covered by the rate cube plane, because background from all wavelengths will fall on all parts of
        a slitless calculation.

        Parameters
        ----------
        rate: 3D numpy.ndarray
            Cube containing the flux rate at the focal plane
        add_extended_background: bool (default: True)
            Toggle for including extended background not contained within the flux cube

        Returns
        -------
        products: 2 entry tuple
            wave_pix: 1D numpy.ndarray containing wavelength to pixel mapping on the detector plane
            spec_rate: 2D numpy.ndarray of detector count rates
        '''
        wave_pix = self.current_instrument.get_wave_pix()
        wave_subs = np.where(
            np.logical_and(
                wave_pix >= self.wave.min(),
                wave_pix <= self.wave.max()
            )
        )
        wave_pix_trunc = wave_pix[wave_subs]

        if len(wave_pix_trunc) == 0:
            raise RangeError(value='wave and wave_pix do not overlap')

        dispersion = self.current_instrument.get_dispersion(wave_pix_trunc)
        trace = self.current_instrument.get_trace(wave_pix_trunc)
        dpix = self.current_instrument.get_detector_pixels(wave_pix_trunc).astype(np.int)

        if dispersion.shape != wave_pix_trunc.shape:
            raise DataConfigurationError("Dispersion and wavelength axes do not match! This should never happen.")

        q_yield, fano_factor = self.current_instrument.get_quantum_yield(wave_pix_trunc)
        # if we kind='slinear' since it's ~2x more memory efficient than 'linear'. 'slinear' uses different code
        # path to calculate the slopes. However, slinear is *much* slower, so it is a tradeoff. Also lowering the
        # rate type to float32 to conserve memory.
        int_rate_pix = sci_int.interp1d(self.wave, rate.astype(np.float32, casting='same_kind'),
                                        kind='linear', axis=2, assume_sorted=True, copy=False)
        rate_pix = int_rate_pix(wave_pix_trunc)

        # convert the photon rate to electron rate by multiplying by the quantum yield which is a function of wavelength
        electron_rate_pix = rate_pix * q_yield

        # to meet IDT expectations, some instruments require a possibly chromatic fudge factor to be applied
        # to the per-pixel electron rate variance.
        var_fudge = self.current_instrument.get_variance_fudge(wave_pix_trunc)

        # the variance in the electron rate, Ve, is also scaled by the quantum yield plus a fano factor which is
        # analytic in the simple 1 or 2 electron case: Ve = (qy + fano) * Re.  since Re is the photon rate
        # scaled by the quantum yield, Re = qy * Rp, we get: Ve = qy * (qy + fano) * Rp
        electron_variance_pix = rate_pix * q_yield * (q_yield + fano_factor) * var_fudge

        # interpolate the background onto the pixel spacing
        int_bg_fp_rate = sci_int.interp1d(self.wave, self.bg_fp_rate.astype(np.float32, casting='same_kind'),
                                          kind='linear', assume_sorted=True, copy=False)
        bg_fp_rate_pix = int_bg_fp_rate(wave_pix_trunc)

        # calculate electron rate and variance due to background
        bg_electron_rate = bg_fp_rate_pix * q_yield
        bg_electron_variance = bg_fp_rate_pix * q_yield * (q_yield + fano_factor) * var_fudge

        # The first part of this code is meant to add the PSF images from the convolved scene cube along either the x
        # or y axis depending on the dispersion axis, optionally following the path of a spectral trace (currently used
        # only for SOSS mode). The psfs will be added to all locations within the resolution element.
        #
        # Because, in slitless modes, the disperser is dispersing light coming in from everywhere in the pupil plane,
        # every part of the detector should have a contribution from every wavelength of light (from both orders, for
        # SOSS mode). The add_extended_background statement does that - it fills every pixel up to i, and after
        # i+rate_pix.shape[1], with the same background that comes baked into the rate_pix images thanks to the
        # AdvancedPSF functions that create the convolved scene cube.

        if self.empty_scene:
            # if we have an explicitly empty scene, we're doing a background-only order calculation and don't need to
            # even pretend to disperse the spectrum - CombinedSignal will handle padding it to match the interesting
            # order(s) and this way there will be no need to trim.
            spec_shape = (rate_pix.shape[0], rate_pix.shape[1])
            spec_rate = np.zeros(spec_shape)
            spec_variance = np.zeros(spec_shape)
            if add_extended_background:
                for i in np.arange(dispersion.shape[0]):
                    spec_rate += bg_electron_rate[i] * dispersion[i]
                    spec_variance += bg_electron_variance[i] * dispersion[i]
        else:  # if the scene is data
            # dispersion_axis tells us whether we need to sum the planes of the cube horizontally
            # or vertically on the detector plane.
            if self.dispersion_axis == 'x':
                spec_shape = (rate_pix.shape[0], rate_pix.shape[2] + rate_pix.shape[1] + dpix[0])
                spec_rate = np.zeros(spec_shape)
                spec_variance = np.zeros(spec_shape)
                for i in range(len(wave_pix_trunc)):
                    # Background not yet completely added. Make sure there is a trace shift to be done so that we
                    # don't make an expensive call to shift() if we don't have to. Use mode='nearest' to fill in new
                    # pixels with background when image is shifted.
                    # dispersion pixel location for any given wavelength is dpix, cross-dispersion pixel location is trace
                    if trace[i] != 0.0:
                        spec_rate[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += shift(
                            electron_rate_pix[:, :, i],
                            shift=(trace[i], 0),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                        spec_variance[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += shift(
                            electron_variance_pix[:, :, i],
                            shift=(trace[i], 0),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                    else:
                        spec_rate[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += electron_rate_pix[:, :, i] * dispersion[i]
                        spec_variance[:, dpix[i]:dpix[i] + rate_pix.shape[1]] += electron_variance_pix[:, :, i] * dispersion[i]

                    # Adding background to all other pixels, unless we are asked not to.
                    if add_extended_background:
                        spec_rate[:, :dpix[i]] += bg_electron_rate[i] * dispersion[i]
                        spec_rate[:, dpix[i] + rate_pix.shape[1]:] += bg_electron_rate[i] * dispersion[i]
                        spec_variance[:, :dpix[i]] += bg_electron_variance[i] * dispersion[i]
                        spec_variance[:, dpix[i] + rate_pix.shape[1]:] += bg_electron_variance[i] * dispersion[i]
            else:  # if the dispersion is on the y axis
                spec_shape = (rate_pix.shape[2] + rate_pix.shape[0] + dpix[0], rate_pix.shape[1])
                spec_rate = np.zeros(spec_shape)
                spec_variance = np.zeros(spec_shape)
                for i in range(len(wave_pix_trunc)):
                    # Background not yet completely added. Make sure there is a trace shift to be done so that we
                    # don't make an expensive call to shift() if we don't have to. Use mode='nearest' to fill in new
                    # pixels with background when image is shifted.
                    # dispersion pixel location for any given wavelength is dpix, cross-dispersion pixel location is trace
                    if trace[i] != 0.0:
                        spec_rate[dpix[i]:dpix[i] + rate_pix.shape[0], :] += shift(
                            electron_rate_pix[:, :, i],
                            shift=(0, trace[i]),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                        spec_variance[dpix[i]:dpix[i] + rate_pix.shape[0], :] += shift(
                            electron_variance_pix[:, :, i],
                            shift=(0, trace[i]),
                            mode='nearest',
                            order=1
                        ) * dispersion[i]
                    else:
                        spec_rate[dpix[i]:dpix[i] + rate_pix.shape[0], :] += electron_rate_pix[:, :, i] * dispersion[i]
                        spec_variance[dpix[i]:dpix[i] + rate_pix.shape[0], :] += electron_variance_pix[:, :, i] * dispersion[i]
                    # Adding background to all other pixels, unless we are asked not to.
                    if add_extended_background:
                        spec_rate[:dpix[i], :] += bg_electron_rate[i] * dispersion[i]
                        spec_rate[dpix[i] + rate_pix.shape[0]:, :] += bg_electron_rate[i] * dispersion[i]
                        spec_variance[:dpix[i], :] += bg_electron_variance[i] * dispersion[i]
                        spec_variance[dpix[i] + rate_pix.shape[0]:, :] += bg_electron_variance[i] * dispersion[i]

        # dispersion_axis determines whether wavelength is the first or second axis
        if self.dispersion_axis == 'x' or self.projection_type == 'multiorder':
            products = wave_pix_trunc, spec_rate, spec_variance
        else:
            # if dispersion is along Y, wavelength increases bottom to top, but Y index increases top to bottom.
            # flip the Y axis to account for this.
            products = wave_pix_trunc, np.flipud(spec_rate), np.flipud(spec_variance)

        return products

    def wave_eff(self, rate):
        rate_tot = np.nansum(rate, axis=0)
        a = np.sum(rate_tot * self.wave)
        b = np.sum(rate_tot)
        if (b > 0.0) and (a > 0.0):
            wave_eff = a / b
        else:
            wave_eff = self.wave.mean()
        wave_eff_arr = np.array([wave_eff])
        return wave_eff_arr

    def get_projection_type(self):
        return self.projection_type

    def ipc_convolve(self, rate, kernel):
        fp_pix_ipc = sg.fftconvolve(rate, kernel, mode='same')

        debug_utils.debugarrays.store('etc3D', 'ipc_convolve',
                                      {
                                          'rate': rate,
                                          'kernel': kernel,
                                          'fp_pix_ipc': fp_pix_ipc,
                                          'description': 'This is just a short, unnecessary description.'
                                      })

        return fp_pix_ipc

    def get_saturation_mask(self, rate=None):
        """
        Compute a numpy array indicating pixels with full saturation (2), partial saturation (1) and no saturation (0).

        Parameters
        ----------
        rate: None or 2D np.ndarray
            Detector plane rate image used to build saturation map from

        Returns
        -------
        mask: 2D np.ndarray
            Saturation mask image
        """
        if rate is None:
            rate = self.rate_plus_bg

        saturation_mask = np.zeros(rate.shape)

        # if (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from det_pars
        if (self.calculation_config.effects['saturation'] if self.calculation_config.effects['saturation'] is not None else self.det_pars['saturation']):
            fullwell = self.det_pars['fullwell']
            exp_pars = self.current_instrument.exposure_spec
            unsat_ngroups = exp_pars.get_unsaturated_groups(rate, fullwell)
            ngroup = exp_pars.ngroup

            saturation_mask[(unsat_ngroups < ngroup)] = 1
            saturation_mask[(unsat_ngroups < 2)] = 2

        return saturation_mask


class CombinedSignal(object):
    """
    This class takes a set of DetectorSignal instances, combines the rates appropriately, and
    provides the other information that DetectorNoise requires.  The primary use for this is in the case
    of SOSS where the detector plane contains signals from effectively three instrument configurations, one
    for each of the visible orders of the gr700xd disperser. These signals need to be combined properly before
    being used to calculate a DetectorNoise and perform a Strategy extraction.

    WARNING: This is currently set up to only support a single aperture slice. Supporting multiple orders with
    multiple slices will require further refactoring...

    Parameters
    ----------
    signal_list: list of DetectorSignal instances
    """

    def __init__(self, signal_list):
        # each signal potentially has a different grid so we want to combine everything onto one grid. instruments
        # that combine multiple signals need to provide detector_pixels so that we know how to shift the signals
        # with respect to each other.
        self.warnings = {}
        self.detector_pixel_list = []
        self.wave_pix_dict = {}
        # some things are common to all signals so get them from the first one
        self.parent_signal = signal_list[0]
        self.signal_list = signal_list
        self.dispersion_axis = self.parent_signal.dispersion_axis

        if len(self.parent_signal.rate_list) > 1:
            msg = "Combining multiple instrument signals onto single detector only supports a single slice."
            raise NotImplementedError(value=msg)

        maxx = 0
        maxy = 0
        for i, s in enumerate(signal_list):
            self.warnings.update(s.warnings)
            pixels = s.detector_pixels
            if pixels is None:
                inst = s.current_instrument.instrument['instrument']
                msg = "No detector pixel configuration set for instrument %s." % inst
                raise DataError(value=msg)
            self.detector_pixel_list.append(pixels)
            ny, nx = s.rate.shape
            maxx = max(maxx, nx)
            maxy = max(maxy, ny)
            self.wave_pix_dict[s.order] = s.wave_pix

        if self.parent_signal.projection_type == "multiorder":
            self.grid = coords.IrregularGrid(np.arange(maxy), np.arange(maxx) - maxx)
        else:
            if self.dispersion_axis == 'x':
                # for slitless calculations, the dispersion axis is longer than the spectrum being dispersed
                # because the whole field of view is being dispersed. 'excess' is the size of the FOV
                # and half will be to the left of the blue end of the spectrum and half to the right of the red end.
                # this is used to create the new spatial coordinate transform for the pixel image on the detector.
                self.grid = coords.IrregularGrid(
                    self.parent_signal.grid.col,
                    (np.arange(maxx) - (maxx) / 2.0) * self.parent_signal.grid.xsamp
                )
            else:
                self.grid = coords.IrregularGrid(
                    (np.arange(maxy) - (maxy) / 2.0) * self.parent_signal.grid.ysamp,
                    self.parent_signal.grid.row
                )

        self.pixgrid_list = [self.grid]
        self.dist = self.grid.dist()
        self.det_mask = np.ones_like(self.dist)

        # these parameters are the same for each signal
        self.wave = self.parent_signal.wave
        self.total_flux = self.parent_signal.total_flux
        self.fp_rate = self.parent_signal.fp_rate
        self.bg_fp_rate = self.parent_signal.bg_fp_rate
        self.background = self.parent_signal.background
        self.flux_cube_list = self.parent_signal.flux_cube_list
        self.flux_plus_bg_list = self.parent_signal.flux_plus_bg_list

        self.aperture_list = self.parent_signal.aperture_list
        self.projection_type = self.parent_signal.projection_type
        self.current_instrument = self.parent_signal.current_instrument
        self.spatial_grid = self.parent_signal.grid
        self.det_pars = self.parent_signal.det_pars
        self.calculation_config = self.parent_signal.calculation_config
        if self.parent_signal.det_pars['rn_correlation']:
            self.read_noise_correlation_matrix = self.parent_signal.read_noise_correlation_matrix

        self.rate_list = [{
            'fp_pix': np.zeros_like(self.dist),
            'fp_pix_no_ipc': np.zeros_like(self.dist),
            'fp_pix_variance': np.zeros_like(self.dist)
        }]
        self.rate_plus_bg_list = [{
            'fp_pix': np.zeros_like(self.dist),
            'fp_pix_no_ipc': np.zeros_like(self.dist),
            'fp_pix_variance': np.zeros_like(self.dist)
        }]
        saturation = np.zeros_like(self.dist)

        for i, s in enumerate(signal_list):
            ny, nx = s.rate_list[0]['fp_pix'].shape
            # set up position within the combined rate image to add each rate
            if self.dispersion_axis == 'x':
                ly = 0
                uy = int(self.rate_list[0]['fp_pix'].shape[0] - s.rate_list[0]['fp_pix'].shape[0])
                # shifts between the rates are referenced to the first pixel location
                lx = 0
                ux = int(self.rate_list[0]['fp_pix'].shape[1] - s.rate_list[0]['fp_pix'].shape[1])
            else:
                # shifts between the rates are referenced to the first pixel location
                ly = 0
                uy = int(self.rate_list[0]['fp_pix'].shape[0] - s.rate_list[0]['fp_pix'].shape[0])
                lx = 0
                ux = int(self.rate_list[0]['fp_pix'].shape[1] - s.rate_list[0]['fp_pix'].shape[1])

            # this will cause pixels that don't overlap to be filled with the background. this is correct for imaging
            # and slitless, but may need revisiting for other cases.
            new_r = np.pad(s.rate_list[0]['fp_pix'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_bg = np.pad(s.rate_plus_bg_list[0]['fp_pix'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_noipc = np.pad(s.rate_list[0]['fp_pix_no_ipc'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_bg_noipc = np.pad(s.rate_plus_bg_list[0]['fp_pix_no_ipc'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_var = np.pad(s.rate_list[0]['fp_pix_variance'], ([ly, uy], [lx, ux]), mode='edge')
            new_r_bg_var = np.pad(s.rate_plus_bg_list[0]['fp_pix_variance'], ([ly, uy], [lx, ux]), mode='edge')
            new_sat = np.pad(s.saturation_list[0], ([ly, uy], [lx, ux]), mode='constant')

            saturation = np.maximum(saturation, new_sat)
            self.rate_list[0]['fp_pix'] += new_r
            self.rate_list[0]['fp_pix_no_ipc'] += new_r_noipc
            self.rate_list[0]['fp_pix_variance'] += new_r_var
            self.rate_plus_bg_list[0]['fp_pix'] += new_r_bg
            self.rate_plus_bg_list[0]['fp_pix_no_ipc'] += new_r_bg_noipc
            self.rate_plus_bg_list[0]['fp_pix_variance'] += new_r_bg_var

        self.saturation_list = [saturation]
        # SOSS only has one aperture so the on_detector rates are just the fp_pix rates.
        self.rate = self.rate_list[0]['fp_pix']
        self.rate_plus_bg = self.rate_plus_bg_list[0]['fp_pix']

        # create ngroup map from the newly combined signal
        self.ngroup_map = signal_list[0].current_instrument.exposure_spec.get_groups_before_sat(self.rate_plus_bg,
                                                                                                self.det_pars[
                                                                                                    'fullwell'])
        self.fraction_saturation = np.max(
            signal_list[0].current_instrument.exposure_spec.get_saturation_fraction(self.rate_plus_bg,
                                                                                    self.det_pars['fullwell']))
        self.brightest_pixel = np.max(self.rate_plus_bg)
        self.groups_list = [self.ngroup_map]

    def get_saturation_mask(self, rate=None):
        """
        Compute a numpy array indicating pixels with full saturation (2), partial saturation (1) and no saturation (0).
        This version just wraps whats implemented within DetectorSignal.

        Parameters
        ----------
        rate: None or 2D np.ndarray
            Detector plane rate image used to build saturation map from

        Returns
        -------
        mask: 2D np.ndarray
            Saturation mask image
        """
        if rate is None:
            rate = self.rate_plus_bg

        mask = self.parent_signal.get_saturation_mask(rate=rate)
        return mask

    def spectral_detector_transform(self):
        """
        Create engine API format dict section containing properties of wavelength coordinates
        at the detector plane.

        Returns
        -------
        t: dict (engine API compliant keys)
        """
        return self.parent_signal.spectral_detector_transform()

    def spectral_model_transform(self):
        """
        Create engine API format dict section containing properties of the wavelength coordinates
        used in the construction of a ConvolvedSceneCube.

        Returns
        -------
        t: dict (engine API compliant keys)
        """
        return self.parent_signal.spectral_model_transform()

    def cube_wcs_info(self):
        """
        Create WCS headers and FITS binary table that describe the cube's coordinate system.

        Returns
        -------
        tbhdu: astropy.io.fits.BinTableHDU instance
            The wavelength sampling is irregularly spaced so we define a binary FITS
            table that contains the array of wavelengths, self.wave.
        header: dict
            Contains the WCS keys that define the coordinate transformation for all axes
        """
        return self.parent_signal.cube_wcs_info()

    def export_to_fits(self, fitsfile='ModelDetectorCube'):
        """
        Write convolved scene cube to a FITS file
        """
        header = self.grid.wcs_info()
        slice_index = 0
        for s,signal in enumerate(self.signal_list):
            for slice_index,flux_plus_bg in enumerate(signal.flux_plus_bg_list):
                fitsfile_slice = '{}_{}_{}.fits'.format(fitsfile, str(slice_index).strip(), s)
                t1 = fits.PrimaryHDU(np.moveaxis(flux_plus_bg, [0, 1, 2], [1, 2, 0]))
                c1 = fits.Column(name='wavelength', array=signal.wave, format='D')
                t2 = fits.BinTableHDU.from_columns([c1])
                t1.header.update(header)
                tbhdu = fits.HDUList([t1,t2])
                tbhdu.writeto(fitsfile_slice, overwrite=True)
                tbhdu.close()


class DetectorNoise(object):
    """
    This class provides the functionality to model the noise in a JWST observation

    Parameters
    ----------
    obs_signal: DetectorSignal or CombinedSignal instance
        The calculated signal at the detector plane
    observation: observation.Observation instance
        Configuration information for the Observation that generated ObsSignal
    """

    def __init__(self, obs_signal, observation):
        self.warnings = {}
        self.observation = observation

        self.grid = obs_signal.grid
        self.dist = obs_signal.dist
        self.det_mask = obs_signal.det_mask
        self.det_pars = obs_signal.det_pars
        self.calculation_config = obs_signal.calculation_config

        self.current_instrument = observation.instrument

        # Are we dealing with an image or spectroscopic mode?
        self.projection_type = self.current_instrument.projection_type

        # get saturation so we can calculate unsaturated groups
        # if (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from det_pars
        if (self.calculation_config.effects['saturation'] if self.calculation_config.effects['saturation'] is not None else self.det_pars['saturation']):
            if 'fullwell' not in self.det_pars:
                msg = "Detector fullwell configuration missing for %s." % self.instrument.instrument
                raise DataError(value=msg)
            self.fullwell = self.det_pars['fullwell']
        else:
            self.fullwell = 1.0e99

        # if not explicitly set, assume we just need 2 groups to define a slope
        self.mingroups = self.det_pars.get("mingroups", 2)

        # if (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from det_pars
        if (self.calculation_config.noise['crs'] if self.calculation_config.noise['crs'] is not None else self.det_pars['crs']):
            # if we are including CRs, we need a CR rate configured for the telescope. this rate is in events/s/cm^2/sr.
            if not hasattr(self.current_instrument.telescope, 'cr_rate'):
                msg = "No cosmic ray rate defined for %s." % self.current_instrument.telescope.tel_name.upper()
                raise DataError(value=msg)
            self.cr_rate = self.current_instrument.telescope.cr_rate

            # we need physical pixel size to calculate incident area in sq cm
            if 'pix_size' not in self.det_pars:
                msg = "The physical pixel size of the detector in microns must be specified to calculate the CR rate."
                raise DataError(value=msg)
            self.pix_size = self.det_pars["pix_size"]

            # need to scale the rate by number of pixels affected by each event.
            if not hasattr(self.current_instrument.telescope, 'cr_npixels'):
                msg = "No defined value for number of pixels affected per CR event for %s." % \
                      self.current_instrument.telescope.tel_name.upper()
                raise DataError(value=msg)
            self.cr_npixels = self.current_instrument.telescope.cr_npixels

            cr_rate = self.cr_rate * self.cr_npixels

            # cr_rate is in units of events/s/cm^2/sr.  convert pix_size in um to cm and multiply by 4pi steradians
            # to get cr_rate per pixel.
            self.pix_cr_rate = (self.pix_size * 1.0e-4) ** 2 * 4.0 * np.pi * cr_rate
        else:
            self.pix_cr_rate = 0.0

        self.var_pix_list, self.stdev_pix_list, self.rn_var_pix_list , self.ff_pix_list, self.dc_var_pix_list = self.basic_source_noise(obs_signal)
        self.var_pix, self.stdev_pix, self.var_rn_pix, self.ff_pix, self.var_dc_pix = self.on_detector()

    def on_detector(self):
        """
        Calculates the detector plane noise products

        Returns
        -------
        products: tuple
            detector_var - numpy.ndarray
                Full variance including all noise sources
            detector_stdev - numpy.ndarray
                Standard deviation (i.e. sqrt(detector_var))
            detector_rn_var - numpy.ndarray
                Variance stricly due to detector readnoise
            detector_ff_var - numpy ndarray
                Variance strictly due to flat field noise
            detector_dc_var - numpy ndarray
                Variance strictly due to dark current
        """
        aperture_sh = self.var_pix_list[0].shape
        n_apertures = len(self.var_pix_list)
        detector_shape = (aperture_sh[0] * n_apertures, aperture_sh[1])
        detector_var = np.zeros(detector_shape)
        detector_stdev = np.zeros(detector_shape)
        detector_rn_var = np.zeros(detector_shape)
        detector_ff = np.zeros(detector_shape)
        detector_dc_var = np.zeros(detector_shape)

        i = 0
        for var_pix, stdev_pix, rn_var_pix, ff_pix, dc_var_pix in zip(self.var_pix_list, self.stdev_pix_list, self.rn_var_pix_list, self.ff_pix_list, self.dc_var_pix_list):
            detector_var[i * aperture_sh[0]:(i + 1) * aperture_sh[0]:] = var_pix
            detector_stdev[i * aperture_sh[0]:(i + 1) * aperture_sh[0]:] = stdev_pix
            detector_rn_var[i * aperture_sh[0]:(i + 1) * aperture_sh[0]:] = rn_var_pix
            detector_ff[i * aperture_sh[0]:(i + 1) * aperture_sh[0]:] = ff_pix
            detector_dc_var[i * aperture_sh[0]:(i + 1) * aperture_sh[0]:] = dc_var_pix

            i += 1

        output = self.det_mask * detector_var, self.det_mask * detector_stdev, self.det_mask * detector_rn_var, self.det_mask * detector_ff, self.det_mask * detector_dc_var
        return output

    def basic_source_noise(self, obs_signal):
        """
        Calculate the noise using the full pixelated flux cube.

        Inputs
        ------
        obs_signal: ObsSignal class instance
            Class containing the detector flux plane or plane cube

        Returns
        -------
        var_pix_list: List of ndarrays
            The detector variance plane or plane cube
        stdev_pix_list: List of ndarrays
            The detector standard deviation plane or plane cube (literally the square root of var_pix_list).
        rn_var_pix_list: List of ndarrays
            Variance strictly due to detector readnoise
        ffnoise_pix_list: List of ndarrays
            Variance strictly due to flat field noise
        dc_var_pix_list: List of ndarrays
            Variance strictly due to dark current    
        """
        exp_pars = self.current_instrument.exposure_spec
        ff_electrons = self.det_pars['ff_electrons']

        var_pix_list = []
        stdev_pix_list = []
        rn_var_pix_list = []
        ffnoise_pix_list = []
        dc_var_pix_list = []

        for rate_plus_bg in obs_signal.rate_plus_bg_list:
            slope_var_pix, slope_rn_var_pix, slope_dc_var_pix = self.get_slope_variance(rate_plus_bg)
            rate_per_pix = rate_plus_bg['fp_pix']
            """
            The flat field error is a division by ~1 (the flat field is normalized), with a variance of 1/ff_electrons.
            Note that the value of the flat field response is constant for multiple ramps and multiple integrations, so
            nramps > 1 does not decrease the residual flat field noise. Due to that, a user will either have to improve the
            flat field or dither with > 1 pixel offsets. The most apparent effect for everyday ETC use is that this sets an
            upper limit on the achievable signal-to-noise ratio.

            The pixel variance upon division with a normalized flat field constructed with ff_electrons (it is assumed that
            the flat field is ideal):

            s^2(R/FF) = s^2(R) + R^2/FF_electrons
            """
            var_pix = slope_var_pix / exp_pars.nramps
            rn_var_pix = slope_rn_var_pix / exp_pars.nramps
            dc_var_pix = slope_dc_var_pix / exp_pars.nramps

            # Add the flat field residual noise if requested. If there are multiple exposures, decrease the noise
            # to account for the effect of co-adding exposures.
            # if (a if condition else b)
            # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from det_pars
            if (self.calculation_config.noise['ffnoise'] if self.calculation_config.noise['ffnoise'] is not None else self.det_pars['ffnoise']):
                ffnoise_pix = (rate_per_pix ** 2 / ff_electrons) / exp_pars.nexp
                var_pix += ffnoise_pix
            else:
                ffnoise_pix = np.zeros_like(rate_per_pix)

            stdev_pix = np.sqrt(var_pix)

            var_pix_list.append(var_pix)
            stdev_pix_list.append(stdev_pix)
            rn_var_pix_list.append(rn_var_pix)
            dc_var_pix_list.append(dc_var_pix)
            ffnoise_pix_list.append(ffnoise_pix)

        products = var_pix_list, stdev_pix_list, rn_var_pix_list, ffnoise_pix_list, dc_var_pix_list
        return products

    def calc_cr_loss(self, ngroups):
        """
        Calculate the effective loss of exposure time due to cosmic rays. This uses the cosmic ray (CR) event rate
        that is contained within the telescope configuration to calculate the odds of a cosmic rays
        hitting a pixel. Pixels that are hit by cosmic rays are assumed to have ramps that are valid before the
        CR, but not after. The exposure specification is used to adjust the expectation value of the exposure
        time (i.e. ngroups) for a CR-truncated ramp.

        See discussion and links in:

        - https://confluence.stsci.edu/pages/viewpage.action?spaceKey=JWST&title=2014-11-10+Effect+of+CRs+on+SNR
        - Robberto (2010) Technical Report JWST-STScI-001928

        for more information on implementation and the numbers used.

        Parameters
        ----------
        ngroups: int
            Number of groups over which to calculate CR losses.  This will usually be number of unsaturated groups.

        Returns
        -------
        cr_ngroups: float
            This is the input ngroups scaled by the mean loss of time due to cosmic ray events.
        """
        # if not (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from det_pars
        if not (self.calculation_config.noise['crs'] if self.calculation_config.noise['crs'] is not None else self.det_pars['crs']):
            # if we're not correcting for CRs, just pass ngroups back
            cr_ngroups = ngroups
        else:
            exp_pars = self.current_instrument.exposure_spec

            # this is the average fraction of the ramp that is lost upon a CR event
            if ngroups < self.mingroups:
                # if for some reason (e.g. using single read noise model) ngroup is less than the minimum needed
                # to get a good ramp fit, then the whole ramp is lost upon a CR event
                ramp_frac = 1.0
            else:
                # the more groups you have per ramp, the less of the ramp you lose to CRs on average.
                # this formalism takes into account that there's always a fraction that's totally lost.
                # in the limit of infinite reads this converges on half the ramp since CRs are evenly
                # distributed.
                ramp_frac = 1.0 - 0.5 * (ngroups - self.mingroups) / (ngroups - 1.0)

            # the effective ramp exposure/measurement time, t_eff, is t_tot * (1 - ramp_frac * pix_cr_rate * t_ramp). We use the saturation
            # time as an approximation of the time during which a ramp can get damaged by a CR. This was previously more complex to
            # handle the case where a ramp is both saturated and hit by a CR, but that required a recalculation of the exposure time for
            # the unsaturated groups. We now simplify this (the difference is minimal and would add some complex logic now that the time
            # formulae are detector type dependent). This is simple.
            cr_ngroups = (1.0 - ramp_frac * self.pix_cr_rate * exp_pars.saturation_time) * ngroups

        return cr_ngroups

    def get_slope_variance(self, rate):
        """
        Compute the slope variance, based on detector properties and
        exposure specifications.

        Inputs
        ------
        rate_per_pix: ndarray
           The measured slope of the ramp (ie, the rate) per pixel
           This can be a one-dimensional wavelength spectrum (for a 1D ETC)
        or a three-dimensional cube ([wave,x,y] for a 3D ETC).

        Returns
        -------
        slope_var: ndarray
           The associated variance of the measured slope
        slope_rn_var: ndarray
           The associated variance of the readnoise only
        slope_dc_var: ndarray
           The associated variance of the dark current only
        """
        dark_current = self.det_pars.get('dark_current', 0.0)
        rn = self.det_pars.get('rn', 0.0)
        var_fudge = self.det_pars.get('var_fudge', 1.0)
        rn_fudge = self.det_pars.get('rn_fudge', 1.0)
        excessp1 = self.det_pars.get('excessp1', 0.0)
        excessp2 = self.det_pars.get('excessp2', 0.0)

        exp_pars = self.current_instrument.exposure_spec

        unsat_ngroups = exp_pars.get_unsaturated_groups(rate['fp_pix_no_ipc'], self.fullwell,
                                                        full_saturation=self.mingroups)
        # scale the unsaturated ngroups by the CR loss.
        #
        # The fact that unsat_ngroups is not an integer is intentional, and is the way the ETC statistically
        # accounts for CR losses.
        # if (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from det_pars
        if (self.calculation_config.noise['crs'] if self.calculation_config.noise['crs'] is not None else self.det_pars['crs']):
            unsat_ngroups = np.vectorize(self.calc_cr_loss)(unsat_ngroups)

        slope_var, slope_rn_var, slope_dc_var = exp_pars.slope_variance(rate, dark_current, rn, unsat_ngroups,
                                                          rn_fudge, excessp1, excessp2)

        return slope_var, slope_rn_var, slope_dc_var

    def get_readnoise_slope_variance(self, rate):
        """
        Compute the variance of the read noise only, excluding all other noise contributions.

        Inputs
        ------
        none

        Returns
        -------
        var_rn: ndarray
            The variance of the readnoise given the current detector slope parameters. Note that
            if the readnoise is switched off, the associated variance will of course be zero.
        """
        rn = self.det_pars.get('rn', 0.0)

        exp_pars = self.current_instrument.exposure_spec
        unsat_ngroups = exp_pars.get_unsaturated_groups(rate['fp_pix_no_ipc'], self.fullwell,
                                                        full_saturation=self.mingroups)
        # scale the unsaturated ngroups by the CR loss.
        # if (a if condition else b)
        # if the calculation_config value is True or False (edited by user), use True or False. If it's none, use the instrument-team-defined default from det_pars
        if (self.calculation_config.noise['crs'] if self.calculation_config.noise['crs'] is not None else self.det_pars['crs']):
            unsat_ngroups = np.vectorize(self.calc_cr_loss)(unsat_ngroups)

        var_rn = exp_pars.rn_variance(rn, unsat_ngroups=unsat_ngroups)

        return var_rn


def calculate_sn(input, webapp=False):
    """
    This is a function to do the 'forward' exposure time calculation where given a dict
    in engine API input format we calculate the resulting Signal/Noise and return a Report
    on the results.

    Parameters
    ----------
    input: dict
        Engine API format dictionary containing the information required to perform the calculation.
    psf_library : psf_library.PSFLibrary instance
        Library of PSF files (e.g. produced by webbpsf) to be used in the calculation

    Returns
    -------
    report.Report instance
    """
    warnings = {}
    try:
        scene_configuration = input['scene']
        background = input['background']
        background_level = input.get('background_level', 'medium')
        instrument_configuration = input['configuration']
        strategy_configuration = input['strategy']
        if input.get('debugarrays'):
            debug_utils.init(input.get('debugarrays'))
    except KeyError as e:
        message = "Missing information required for the calculation: %s" % str(e)
        raise EngineInputError(value=message)

    # get the calculation configuration from the input or use the defaults
    if 'calculation' in input:
        calc_config = CalculationConfig(config=input['calculation'])
    else:
        calc_config = CalculationConfig()

    # #### BEGIN calculation #### #
    """
    This section currently implements the Pandeia engine API.  As the engine's object model
    is refactored, this section will have to change accordingly.
    """
    # check for empty scene configuration and set it up properly if it is empty.
    if len(scene_configuration) == 0:
        scene_configuration = build_empty_scene()

    scene = Scene(input=scene_configuration, webapp=webapp)
    warnings.update(scene.warnings)
    instrument = InstrumentFactory(config=instrument_configuration, webapp=webapp)
    warnings.update(instrument.warnings)

    strategy = StrategyFactory(instrument, config=strategy_configuration, webapp=webapp)

    # strategies can have different figures of interest that need to be calculated.
    # in most cases, S/N is what is desired. However, for coronagraphy the figure of interest
    # is sometimes the contrast that can be achieved. In this case, strategy.calc_type will be
    # set to 'contrast' and we need to run calculate_contrast.
    if hasattr(strategy, "calc_type"):
        if strategy.calc_type == "contrast":
            r = calculate_contrast(input, webapp=webapp)
        else:
            msg = "Unsupported calculation type: %s" % strategy.calc_type
            raise EngineInputError(value=msg)
    else:
        # set up the observation and then do S/N calculation...
        obs = observation.Observation(
            scene=scene,
            instrument=instrument,
            strategy=strategy,
            background=background,
            background_level=background_level,
            webapp=webapp
        )

        # seed the random number generator
        seed = obs.get_random_seed()
        np.random.seed(seed=seed)

        # Sometimes there is more than one exposure involved so implement lists for signal and noise
        my_detector_signal_list = []
        my_detector_noise_list = []
        my_detector_saturation_list = []

        if hasattr(strategy, 'dithers'):
            dither_list = strategy.dithers
        else:
            dither_list = [{'x': 0.0, 'y': 0.0}]

        for dither in dither_list:
            o = copy.deepcopy(obs)
            o.scene.offset(dither)
            # Calculate the signal rate in the detector plane. If they're configured, need to loop through
            # configured orders to include all dispersed signal.
            if instrument.projection_type in ['slitless', 'multiorder']:
                if 'orders' in instrument.disperser_config[instrument.instrument['disperser']]:
                    if instrument.instrument['filter'] in instrument.disperser_config[instrument.instrument['disperser']]['orders']:
                        orders = instrument.disperser_config[instrument.instrument['disperser']]['orders'][instrument.instrument['filter']]
                    else:
                        orders = instrument.disperser_config[instrument.instrument['disperser']]['orders']
                else:
                    orders = None
                if 'background_orders' in instrument.disperser_config[instrument.instrument['disperser']]:
                    if instrument.instrument['filter'] in instrument.disperser_config[instrument.instrument['disperser']]['background_orders']:
                        background_orders = instrument.disperser_config[instrument.instrument['disperser']]['background_orders'][instrument.instrument['filter']]
                    else:
                        background_orders = instrument.disperser_config[instrument.instrument['disperser']]['background_orders']
                    background_orders = instrument.disperser_config[instrument.instrument['disperser']][
                        'background_orders']
                else:
                    background_orders = []
            else:
                orders = None

            if orders is not None:
                order_signals = []
                # First, take care of the orders for which a signal is desired.
                for order in orders:
                    order_signals.append(DetectorSignal(o, calc_config=calc_config, webapp=webapp, order=order))
                # Then loop through a list of orders for which we only want to add the background calculation
                if background_orders is not None:
                    o_empty = copy.deepcopy(obs)
                    o_empty.scene = Scene(input=build_empty_scene(), webapp=webapp)
                    for order in background_orders:
                        order_signals.append(
                            DetectorSignal(o_empty, calc_config=calc_config, webapp=webapp, order=order,
                                           empty_scene=True))

                my_detector_signal = CombinedSignal(order_signals)
            else:
                my_detector_signal = DetectorSignal(o, calc_config=calc_config, webapp=webapp, order=None)

            my_detector_noise = DetectorNoise(my_detector_signal, o)

            # Every dither has a saturation map
            my_detector_saturation = my_detector_signal.get_saturation_mask()

            my_detector_signal_list.append(my_detector_signal)
            my_detector_noise_list.append(my_detector_noise)
            my_detector_saturation_list.append(my_detector_saturation)

        # Use the strategy to get the extracted signal/noise products
        extracted_sn_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)
        warnings.update(extracted_sn_list[0]['warnings'])
        # #### END calculation #### #
        r = ReportFactory(input, my_detector_signal_list, my_detector_noise_list, my_detector_saturation_list,
                          extracted_sn_list, warnings)

    return r


def calculate_contrast(input, webapp=False):
    """
    This is a function to do the 'forward' exposure time calculation where given a dict
    in engine API input format we calculate the resulting coronagraphic contrast and return a Report
    on the results.

    While this method is meant for coronagraphic modes, it will work also for regular imaging modes.

    Parameters
    ----------
    input: dict
        Engine API format dictionary containing the information required to perform the calculation.
    psf_library : psf_library.PSFLibrary instance
        Library of PSF files (e.g. produced by webbpsf) to be used in the calculation

    Returns
    -------
    report.Report instance
    """
    warnings = {}
    try:
        scene_configuration = input['scene']
        background = input['background']
        background_level = input.get('background_level', 'medium')
        instrument_configuration = input['configuration']
        strategy_configuration = input['strategy']
        psf_subtraction_configuration = input['strategy']['psf_subtraction_source']
        if input.get('debugarrays'):
            debug_utils.init(input.get('debugarrays'))
    except KeyError as e:
        message = "Missing information required for the calculation: %s" % str(e)
        raise EngineInputError(value=message)

    # get the calculation configuration from the input or use the defaults
    if 'calculation' in input:
        calc_config = CalculationConfig(config=input['calculation'])
    else:
        calc_config = CalculationConfig()

    # #### BEGIN calculation #### #
    """
    This section implements the Pandeia engine API.
    """
    # check for empty scene configuration and set it up properly if it is empty.
    if len(scene_configuration) == 0:
        scene_configuration = build_empty_scene()

    instrument = InstrumentFactory(config=instrument_configuration, webapp=webapp)
    warnings.update(instrument.warnings)

    # move the psf_reference to a pre-determined and fixed location outside of the FOV
    strategy = StrategyFactory(instrument, config=strategy_configuration, webapp=webapp)
    # Dither weight settings for Coronagraphy are now controlled by a do_contrast attribute of Strategy.
    # Set this to False to get the coronagraphy PSF subtraction, set to True to get the unocculted dither.
    strategy.do_contrast = False

    # Check for user-specified dithers (the contrast calculation will add an additional 2 fictional dithers).
    if not hasattr(strategy, 'dithers') or len(strategy.dithers) != 1:
        message = "Contrast calculations currently require a single dither " \
                  "to be passed in the strategy, {} was passed".format(strategy.dithers)
        raise EngineInputError(value=message)

    psf_subtraction_xy = strategy.psf_subtraction_xy
    pointing_error = strategy.pointing_error
    psf_subtraction_configuration['position']['x_offset'] = strategy.psf_subtraction_xy[0]
    psf_subtraction_configuration['position']['y_offset'] = strategy.psf_subtraction_xy[1]

    # add the psf_reference to the scene
    scene_configuration.append(psf_subtraction_configuration)
    scene = Scene(input=scene_configuration, webapp=webapp)
    if hasattr(strategy, "scene_rotation"):
        scene.sources = strategy.rotate(scene.sources)
    warnings.update(scene.warnings)
    # add an attribute specifying the PSF subtraction source
    scene.psf_subtraction_idx = len(scene_configuration) - 1

    # Add the appropriate dithers for the PSF reference star and the unocculted dither
    psf_subtraction_dither = {
        'x': -strategy.psf_subtraction_xy[0] - strategy.pointing_error[0],
        'y': -strategy.psf_subtraction_xy[1] - strategy.pointing_error[1]
    }
    unocculted_dither = {'x': strategy.unocculted_xy[0], 'y': strategy.unocculted_xy[1]}

    strategy.dithers.append(psf_subtraction_dither)
    strategy.dithers.append(unocculted_dither)
    strategy.on_target = [True, False, False]

    # set up the observation...
    obs = observation.Observation(
        scene=scene,
        instrument=instrument,
        strategy=strategy,
        background=background,
        background_level=background_level,
        webapp=webapp
    )

    # seed the random number generator
    seed = obs.get_random_seed()
    np.random.seed(seed=seed)

    # Sometimes there is more than one exposure involved so implement lists for signal and noise
    my_detector_signal_list = []
    my_detector_noise_list = []
    my_detector_saturation_list = []

    for dither in strategy.dithers:
        # make a new deep copy of the observation for each dither so that each position is offset
        # from the center position. otherwise the offsets get applied cumulatively via the reference.
        o = copy.deepcopy(obs)
        o.scene.offset(dither)
        # Calculate the signal rate in the detector plane
        my_detector_signal = DetectorSignal(o, calc_config=calc_config, webapp=webapp)
        my_detector_noise = DetectorNoise(my_detector_signal, o)

        # Every dither has a saturation map
        my_detector_saturation = my_detector_signal.get_saturation_mask()
        my_detector_signal_list.append(my_detector_signal)
        my_detector_noise_list.append(my_detector_noise)
        my_detector_saturation_list.append(my_detector_saturation)

    # We need a regular S/N of the target source
    extracted_sn_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)
    warnings.update(extracted_sn_list[0]['warnings'])

    # Use the strategy to get the extracted contrast products
    grid = my_detector_signal_list[0].grid

    aperture = strategy.aperture_size
    annulus = strategy.sky_annulus

    # Create a list of contrast separations for which to calculate the contrast
    bounds = grid.bounds()
    ncontrast = strategy.ncontrast
    contrasts = np.zeros(ncontrast)
    contrast_separations = np.linspace(0 + aperture, bounds['xmax'] - annulus[1], ncontrast)
    contrast_azimuth = np.radians(strategy.contrast_azimuth)
    contrast_xys = [(separation * np.sin(contrast_azimuth),
                     separation * np.cos(contrast_azimuth)) for separation in contrast_separations]

    # Calculate contrast at each separation
    for i, contrast_xy in enumerate(contrast_xys):
        strategy.target_xy = contrast_xy
        extracted_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)
        contrasts[i] = extracted_list[0]['extracted_noise']

    # What is the flux of the unocculted star.
    # We set the do_contrast attribute to True so the unocculted dither will be used.
    strategy.do_contrast = True
    strategy.on_target = [False, False, True]
    strategy.target_xy = strategy.unocculted_xy
    extract_unocculted_list = strategy.extract(my_detector_signal_list, my_detector_noise_list)
    extracted_sn_list.append(extract_unocculted_list[0])

    # Contrast is relative to the unocculted on-axis star.
    contrasts /= extract_unocculted_list[0]['extracted_flux']
    contrast_curve = [contrast_separations, contrasts]

    # #### END calculation #### #

    # Add the contrast curve and link relevant saturation maps to the extracted_sn dict for passing
    for x in range(len(extracted_sn_list)):
        extracted_sn_list[x]['contrast_curve'] = contrast_curve

    r = ReportFactory(input, my_detector_signal_list, my_detector_noise_list, my_detector_saturation_list,
                      extracted_sn_list, warnings)
    return r


def calculate_exposure_time(input, webapp=False, **kwargs):
    """
    This is a function to do the 'reverse' exposure time calculation where given a desired
    signal-to-noise ratio we calculate the optimal exposureSpecification and return a Report
    on the results.

    Parameters
    ----------
    input: dict
        Dictionary containing the information required to perform the calculation.
    psf_library : psf_library.PSFLibrary instance
        Library of PSF files (e.g. produced by webbpsf) to be used in the calculation

    Returns
    -------
    report.Report instance
    """
    raise NotImplementedError("Reverse ETC calculations are not yet implemented.")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Module that implements a background noise calculator (BNC) for the pandeia ETC engine
"""

from __future__ import division, absolute_import

import numpy as np
from warnings import warn

from .observation import Observation
from .scene import Scene
from .custom_exceptions import EngineInputError
from .instrument_factory import InstrumentFactory
from .strategy import StrategyFactory
from .calc_utils import build_empty_scene
from .etc3D import DetectorSignal, DetectorNoise, CombinedSignal, CalculationConfig
from . import debug_utils


def bnc(calc_input):
    """
    This function uses the pandeia ETC engine to calculate the standard deviation of the per-pixel count rate for a
    given calculation when only the background signal is included. The input can be any valid pandeia calculation.
    The scene information will be ignored and the strategy information ignored except for number of dithers. The output
    will be the maximum standard deviation (sqrt(variance)) encountered at the detector plane.

    Parameters
    ----------
    calc_input: dict
        Pandeia engine input API compliant dict containing information required to perform the background noise calculation

    Returns
    -------
    pix_stddev: float
        Maximum standard deviation in pixel count rate (e-/sec/pixel) generated by the input background signal.
    """

    # ignore the scene in the input calculation and use an empty one instead. the only input signal in the calculation will
    # be the background which is spatially constant across the scene.
    scene_configuration = build_empty_scene()

    debug_utils.init(calc_input)

    try:
        background = calc_input['background']
        background_level = calc_input.get('background_level', 'medium')
        instrument_configuration = calc_input['configuration']
        # we need to pass a strategy to Observation and we also need to know the number of dithers to
        # calculate the background variance in the combined detector plane image
        strategy_configuration = calc_input['strategy']
        if calc_input.get('debugarrays'):
            debug_utils.init(calc_input.get('debugarrays'))
    except KeyError as e:
        message = "Missing information required for the calculation: %s" % str(e)
        raise EngineInputError(value=message)

    # get the calculation configuration from the input or use the defaults
    if 'calculation' in calc_input:
        calc_config = CalculationConfig(config=calc_input['calculation'])
    else:
        calc_config = CalculationConfig()

    scene = Scene(input=scene_configuration)
    instrument = InstrumentFactory(config=instrument_configuration)
    strategy = StrategyFactory(instrument, config=strategy_configuration)

    obs = Observation(
        scene=scene,
        instrument=instrument,
        strategy=strategy,
        background=background,
        background_level=background_level
    )

    # seed the random number generator
    seed = obs.get_random_seed()
    np.random.seed(seed=seed)

    # dithering images uncorrelates all of the noise and thus reduces the variance by ndithers:
    #   variance = variance_single/ndithers
    if hasattr(strategy, 'dithers'):
        ndithers = len(strategy.dithers)
    else:
        ndithers = 1

    # Calculate the signal rate in the detector plane. If they're configured, need to loop through
    # configured orders to include all dispersed signal.
    if instrument.projection_type in ['slitless', 'multiorder']:
        if 'orders' in instrument.disperser_config[instrument.instrument['disperser']]:
            if instrument.instrument['filter'] in instrument.disperser_config[instrument.instrument['disperser']]['orders']:
                orders = instrument.disperser_config[instrument.instrument['disperser']]['orders'][instrument.instrument['filter']]
            else:
                orders = instrument.disperser_config[instrument.instrument['disperser']]['orders']
            # the point of background orders in normal operation is that they use 
            # an empty scene, and just provide additional background flux. 
            # However, that is the entire point of bnc.py - it uses the input 
            # parameters and an empty scene. So orders and background orders do 
            # not need to be separate lists or loops.
            if 'background_orders' in instrument.disperser_config[instrument.instrument['disperser']]:
                if instrument.instrument['filter'] in instrument.disperser_config[instrument.instrument['disperser']]['background_orders']:
                    orders.extend(instrument.disperser_config[instrument.instrument['disperser']]['background_orders'][instrument.instrument['filter']])
                else:
                    orders.extend(instrument.disperser_config[instrument.instrument['disperser']]['background_orders'])
        else:
            orders = None
    else:
        orders = None

    if orders is not None:
        order_signals = []
        for order in orders:
            order_signals.append(DetectorSignal(obs, calc_config=calc_config, order=order))
        signal = CombinedSignal(order_signals)
    else:
        signal = DetectorSignal(obs, calc_config=calc_config)

    noise = DetectorNoise(signal, obs)

    # get the detector plane noise products
    det_var, det_stddev, det_rn_var, det_ff_var, det_dc_var = noise.on_detector()

    pix_stddev = det_stddev.max() / np.sqrt(ndithers)
    return pix_stddev



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# This is the old WFIRST iPython Notebook GUI - it should be replaced by the new notebook utils file in
# nb_utils2.py, because this was written for an old version of iPyWidgets and (while it still produces results)
# it doesn't behave correctly and the look is correspondingly rather cluttered and confusing.

from __future__ import division, absolute_import

import os
import numpy as np

import matplotlib
from matplotlib import style
style.use('ggplot')
matplotlib.use('nbagg')
import matplotlib.pyplot as plt

from pandeia.engine.perform_calculation import perform_calculation
from pandeia.engine.io_utils import read_json, write_json
from pandeia.engine.calc_utils import build_default_calc

from ipywidgets import widgets
from IPython.display import display, clear_output
import traitlets

def get_config(filename):
    """
    read configuration data from a JSON file and create a dict using display_strings as the keys.
    these dicts will be used to populate pull-downs with descriptive names, but provide lookup to
    the values that need to be passed back to the engine.

    Parameters
    ----------
    filename: string
        filename of JSON config file to Load

    Returns
    -------
    conf: dict
        JSON configuration data with keys swapped out for display_strings where available.
    """
    conf_data = read_json(filename)
    conf = {}
    for k in conf_data:
        if "display_string" in conf_data[k]:
            conf[conf_data[k]["display_string"]] = k
        else:
            conf[k] = k
    return conf

class WFIRST_gui(object):
    """
    create a basic GUI for WFIRST calculations using ipython widgets. to be run from within a jupyter notebook.
    """
    def __init__(self):
        self.r = {}
        self.form = widgets.VBox(width="100%", background_color="#EEE")
        self.src_form = widgets.HBox(padding='10px', width="100%", visibility=False)
        self.src_select = widgets.Dropdown(description="Source type: ", options=['point', 'extended'], value='point')
        self.l1 = widgets.HTML(value="Scale length (arcsec): ", margin="5px")
        self.l2 = widgets.HTML(value="Position angle (deg): ", margin="5px")
        self.l3 = widgets.HTML(value="Ellipticity: ", margin="5px")
        self.sersic = widgets.Dropdown(description="Profile: ", options=["Gaussian", "Exponential", "de Vaucouleurs"])
        self.sersic_idx = {
            "Gaussian": 0.5,
            "Exponential": 1.0,
            "de Vaucouleurs": 4.0
        }
        self.ext_scale = widgets.BoundedFloatText(value=0.2, min=0.0, max=999999.0, width=70)
        self.ellip = widgets.BoundedFloatText(value=0.0, min=0.0, max=1.0, width=70)
        self.posang = widgets.BoundedFloatText(min=0.0, max=360.0, value=0.0, width=70)
        self.src_form.children = [self.sersic, self.l1, self.ext_scale, self.l3, self.ellip, self.l2, self.posang]
        self.src_select.on_trait_change(self.on_src_select, 'value')

        norm_form = widgets.HBox(padding='10px', width="100%")
        self.flux = widgets.BoundedFloatText(description="Source flux: ", min=0.0, max=1.0e30, value=1.0)
        self.units = widgets.Dropdown(value='ujy', options=['abmag', 'njy', 'ujy', 'mjy', 'jy'])
        atwave = widgets.HTML(value=" at ", margin='5px')
        self.wave = widgets.BoundedFloatText(min=0.1, max=99999.0, value=1.5)
        waveunits = widgets.HTML(value="microns", margin='5px')
        norm_form.children = [self.flux, self.units, atwave, self.wave, waveunits]
        self.units.on_trait_change(self.on_units_select, 'value')

        sed_form = widgets.HBox(padding='10px', width="100%")
        self.sed_select = widgets.Dropdown(
            description="SED type: ",
            options=['power-law', 'blackbody', 'star', 'extragalactic'],
            value='power-law'
        )
        self.pl_index = widgets.FloatText(description="Index: ", value=1.0, visibility=True, width=50)
        self.bb_temp = widgets.BoundedFloatText(description="Temp (K): ", min=0.0, max=99999.0, value=6500.0, visibility=False, width=75)

        star_config_file = os.path.join(os.environ['pandeia_refdata'], 'sed', 'phoenix', 'spectra.json')
        self.star_config = get_config(star_config_file)
        self.stars = widgets.Dropdown(options=sorted(self.star_config.keys()), visibility=False)

        gal_config_file = os.path.join(os.environ['pandeia_refdata'], 'sed', 'brown', 'spectra.json')
        self.gal_config = get_config(gal_config_file)
        self.galaxies = widgets.Dropdown(options=sorted(self.gal_config.keys()), visibility=False)

        self.redshift = widgets.BoundedFloatText(description="Redshift:", min=0.0, max=99999.0, value=0.0, width=70)

        self.sed_select.on_trait_change(self.on_sed_select, 'value')
        sed_form.children = [self.sed_select, self.pl_index, self.bb_temp, self.stars, self.galaxies, self.redshift]

        imager_config = read_json(os.path.join(os.environ['pandeia_refdata'], 'wfirst', 'wfirstimager', 'config.json'))
        ifu_config = read_json(os.path.join(os.environ['pandeia_refdata'], 'wfirst', 'wfirstifu', 'config.json'))

        inst_form = widgets.HBox(padding='10px', width="100%")
        self.inst_select = widgets.Dropdown(description="Instrument: ", options=['Imager'], value='Imager')
        im_filters = imager_config['filters']
        im_readout_patterns = imager_config['readout_patterns']
        im_subarrays = imager_config['subarrays']
        self.filt = widgets.Dropdown(description="Filter:", options=im_filters)
        self.readout_pattern = widgets.Dropdown(description="readout_pattern:", options=im_readout_patterns)
        self.subarray = widgets.Dropdown(description="Sub-array:", options=im_subarrays)
        inst_form.children = [self.inst_select, self.filt, self.readout_pattern, self.subarray]

        det_form = widgets.HBox(padding='10px', width="100%")
        self.ngroups = widgets.BoundedIntText(description="Groups: ", min=2, max=999, value=10, width=50)
        self.nints = widgets.BoundedIntText(description="Integrations: ", min=1, max=999, value=1, width=50)
        self.nexps = widgets.BoundedIntText(description="Exposures: ", min=1, max=999, value=1, width=50)
        det_form.children = [self.ngroups, self.nints, self.nexps]

        strat_form = widgets.VBox(padding='10px')
        ap_lab = widgets.HTML(value="Aperture radius (arcsec): ", margin='5px')
        self.ap_size = widgets.BoundedFloatText(min=0.0, max=999.0, value=0.1, width=60)
        self.ap_size.on_trait_change(self.check_ann, 'value')
        self.overplot = widgets.Checkbox(description="Overlay", value=True)
        self.overplot.on_trait_change(self.update_plots)
        hb1 = widgets.HBox(padding='10px', width="100%", children=[ap_lab, self.ap_size, self.overplot])
        bg_lab = widgets.HTML(value="Background annulus radii (arcsec): ", margin='5px')
        self.ann_inner = widgets.BoundedFloatText(description="inner", min=0.0, max=999.0, value=0.2, width="100%")
        self.ann_inner.on_trait_change(self.check_ann_inner, 'value')
        self.ann_outer = widgets.BoundedFloatText(description="outer", min=0.0, max=999.0, value=0.3, width="100%")
        self.ann_outer.on_trait_change(self.check_ann_outer, 'value')
        hb2 = widgets.HBox(padding='10px', width="100%", children=[bg_lab, self.ann_inner, self.ann_outer])
        strat_form.children = [hb1, hb2]

        self.calc_button = widgets.Button(description='Calculate', width="100%", background_color="#bee2c4")
        self.calc_button.on_click(self.run_calc)

        self.plot_form = widgets.HBox(padding='10px', width="100%", pack='center')
        self.oned_plots = {
            "Input Source Flux (mJy)": "target",
            "Input Background (MJy/sr)": "bg",
            "Focal Plane Rate (e-/sec/pixel)": "fp"
        }
        self.oned_units = {
            "target": "mJy",
            "bg": "MJy/sr",
            "fp": "e-/sec/pixel"
        }
        self.twod_plots = {
            "Detector (e-/sec)": "detector",
            "S/N": "snr",
            "Saturation": "saturation"
        }
        self.twod_units = {
            "detector": "e-/sec",
            "snr": "S/N",
            "saturation": ""
        }
        self.oned_pulldown = widgets.Dropdown(
            description="1D Plot",
            options=sorted(self.oned_plots.keys()),
            value="Input Source Flux (mJy)"
        )
        self.twod_pulldown = widgets.Dropdown(
            description="2D Image",
            options=sorted(self.twod_plots.keys()),
            value="S/N"
        )

        self.plot_form.children = [self.oned_pulldown, self.twod_pulldown]
        self.plot_form.visibility = False
        self.oned_pulldown.on_trait_change(self.update_plots)
        self.twod_pulldown.on_trait_change(self.update_plots)

        tlab1 = widgets.HTML(value="<b>Extracted S/N: <b>", margin='5px')
        self.esn = widgets.HTML(value="0.0", margin='5px')
        tlab2 = widgets.HTML(value="       <b>Extracted Flux (e-/sec): </b>", margin='5px')
        self.eflux = widgets.HTML(value="0.0", margin='5px')
        tlab3 = widgets.HTML(value="       <b>Exposure Time (sec): <b>", margin='5px')
        self.etime = widgets.HTML(value="0.0", margin='5px')

        self.tab_form = widgets.HBox(padding='10px', width="100%", pack='center')
        self.tab_form.children = [tlab1, self.esn, tlab2, self.eflux, tlab3, self.etime]
        self.tab_form.visibility = False
        self.form.children = [
            self.src_select,
            self.src_form,
            norm_form,
            sed_form,
            inst_form,
            det_form,
            strat_form,
            self.calc_button,
            self.tab_form,
            self.plot_form
        ]

    def update_plots(self):
        """
        update the 1D and 2D plots.  they're part of the same figure so have to be drawn together.  hard to do
        two independent plots in one cell in a notebook.
        """
        oned_key = self.oned_plots[self.oned_pulldown.value]
        twod_key = self.twod_plots[self.twod_pulldown.value]
        oned_curve = self.r['1d'][oned_key]
        twod_im = self.r['2d'][twod_key]
        fig = plt.figure(figsize=(10, 5))
        ax1 = fig.add_subplot(121)
        plot = ax1.plot(oned_curve[0], oned_curve[1])
        ax1.set_xlabel(r'$\mu m$')
        ax1.set_ylabel(self.oned_units[oned_key])
        t = self.r['transform']
        xmin = t['x_min']
        xmax = t['x_max']
        ymin = t['y_min']
        ymax = t['y_max']
        extent = [xmin, xmax, ymin, ymax]
        ax2 = fig.add_subplot(122)
        ax2.set_xlabel("arcsec")
        ax2.set_ylabel("arcsec", rotation=270)
        ax2.yaxis.tick_right()
        ax2.yaxis.set_label_position("right")
        ax2.yaxis.set_ticks_position("both")
        im = plt.imshow(twod_im, interpolation='nearest', extent=extent)
        if self.overplot.value is True:
            circles = []
            circles.append(plt.Circle((0, 0), radius=self.ap_size.value, edgecolor='white', facecolor='none'))
            circles.append(plt.Circle((0, 0), radius=self.ann_inner.value, edgecolor='red', facecolor='none'))
            circles.append(plt.Circle((0, 0), radius=self.ann_outer.value, edgecolor='red', facecolor='none'))
            for c in circles:
                im.axes.add_artist(c)
        if twod_key == "saturation":
            norm = matplotlib.colors.Normalize(vmin=0, vmax=2)
            im.set_norm(norm)
            c = plt.colorbar(ax=ax2, orientation="horizontal", label=self.twod_units[twod_key], ticks=[0, 1, 2])
            c.ax.set_xticklabels(["None", "Partial", "Full"])
        else:
            c = plt.colorbar(ax=ax2, orientation="horizontal", label=self.twod_units[twod_key])
        plt.tight_layout()
        plt.show()
        clr = clear_output(wait=True)

    @property
    def display(self):
        """
        display the GUI
        """
        display(self.form)

    @property
    def calc_results(self):
        """
        return the calculation results
        """
        return self.r

    @calc_results.setter
    def calc_results(self, r):
        self.r = r
        self.update_plots()

    def check_ann(self, name, value):
        """
        check the background estimation annulus to make sure it's valid

        Parameters
        ----------
        name: string
            not used, but expected for on_trait_change callbacks
        """
        self.check_ann_inner(name=name, value=value)
        self.check_ann_outer(name=name, value=value)

    def check_ann_inner(self, name, value):
        if self.ann_inner.value <= self.ap_size.value:
            self.ann_inner.value = round(self.ap_size.value + 0.1, 3)
        self.check_ann_outer(name=name, value=value)

    def check_ann_outer(self, name, value):
        if self.ann_outer.value - self.ann_inner.value <= 0.1:
            self.ann_outer.value = round(self.ann_inner.value + 0.1, 3)

    def on_src_select(self, name, value):
        if value == 'point':
            self.src_form.visibility = False
        else:
            self.src_form.visibility = True

    def on_units_select(self, name, value):
        if value == 'abmag':
            self.flux.value = 25.0
        else:
            self.flux.value = 1.0

    def on_sed_select(self, name, value):
        if value == 'power-law':
            self.pl_index.visibility = True
            self.bb_temp.visibility = False
            self.stars.visibility = False
            self.galaxies.visibility = False
        elif value == 'blackbody':
            self.pl_index.visibility = False
            self.bb_temp.visibility = True
            self.stars.visibility = False
            self.galaxies.visibility = False
        elif value == 'star':
            self.pl_index.visibility = False
            self.bb_temp.visibility = False
            self.stars.visibility = True
            self.galaxies.visibility = False
        elif value == 'extragalactic':
            self.pl_index.visibility = False
            self.bb_temp.visibility = False
            self.stars.visibility = False
            self.galaxies.visibility = True
        else:
            self.pl_index.visibility = False
            self.bb_temp.visibility = False
            self.stars.visibility = False
            self.galaxies.visibility = False

    def run_calc(self, b):
        c = build_default_calc("wfirst", "wfirstimager", "imaging")
        c['configuration']['detector']['nexp'] = self.nexps.value
        c['configuration']['detector']['ngroup'] = self.ngroups.value
        c['configuration']['detector']['nint'] = self.nints.value
        c['configuration']['detector']['readout_pattern'] = self.readout_pattern.value
        c['configuration']['detector']['subarray'] = self.subarray.value
        c['configuration']['instrument']['filter'] = self.filt.value

        src = c['scene'][0]
        if self.src_select.value == "extended":
            src['shape']['geometry'] = 'sersic'
            a = self.ext_scale.value
            e = self.ellip.value
            b = (1.0 - e) * a
            s_idx = self.sersic_idx[self.sersic.value]
            # if gaussian, convert a/b to sigma
            if s_idx == 0.5:
                a *= np.sqrt(2.0)
                b *= np.sqrt(2.0)
            src['shape']['major'] = a
            src['shape']['minor'] = b
            src['shape']['sersic_index'] = s_idx
            src['position']['orientation'] = self.posang.value

        src['spectrum']['redshift'] = self.redshift.value
        src['spectrum']['normalization']['norm_flux'] = self.flux.value
        src['spectrum']['normalization']['norm_fluxunit'] = self.units.value
        src['spectrum']['normalization']['norm_wave'] = self.wave.value

        sed = self.sed_select.value
        if sed == "power-law":
            src['spectrum']['sed']['sed_type'] = "powerlaw"
            src['spectrum']['sed']['index'] = self.pl_index.value
        if sed == "blackbody":
            src['spectrum']['sed']['sed_type'] = "blackbody"
            src['spectrum']['sed']['temp'] = self.bb_temp.value
        if sed == "star":
            src['spectrum']['sed']['sed_type'] = "phoenix"
            src['spectrum']['sed']['key'] = self.star_config[self.stars.value]
        if sed == "extragalactic":
            src['spectrum']['sed']['sed_type'] = "brown"
            src['spectrum']['sed']['key'] = self.gal_config[self.galaxies.value]

        c['strategy']['aperture_size'] = self.ap_size.value
        ann = [self.ann_inner.value, self.ann_outer.value]
        c['strategy']['sky_annulus'] = ann

        self.r = perform_calculation(c, dict_report=True)
        self.calc_input = c
        self.plot_form.visibility = True
        self.esn.value = "%.2f" % self.r['scalar']['sn']
        self.eflux.value = "%.2f" % self.r['scalar']['extracted_flux']
        self.etime.value = "%.2f" % self.r['scalar']['total_exposure_time']
        self.tab_form.visibility = True

        self.update_plots()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import os
import copy
import astropy
import pkg_resources
from . import io_utils as io
from .utils import merge_data
from .pandeia_warnings import standard_warning_messages as warning_messages


default_refdata_directory = os.environ.get("pandeia_refdata")


def default_refdata(directory=None):
    """
    There is a default refdata that is either refdata/ (for convenience
    of a developer) or the value of the environment variable $pandeia_refdata
    After importing this module, you can call this function to explicitly
    set the name of the refdata directory
    """
    global default_refdata_directory
    if directory is not None:
        default_refdata_directory = directory
    return default_refdata_directory


class DefaultConfig(object):

    """
    This class provides functionality to discover and load defaults from JSON
    configuration files, update the configuration information from a dict and any kwargs
    passed from the caller, and then populate the objects attributes from the resulting dict.

    Parameters
    ----------
    config: dict (optional)
        dictionary containing necessary configuration information
    **kwargs: list of keyword/value pairs
        parameter keyword and value pairs to augment defaults and config
    """

    def __init__(self, config={}, webapp=False, **kwargs):
        # grab info from the configured defaults file, if any, from caller via a passed dict, or via keywords
        # clean meta blocks from the configuration files
        all_config = merge_data(self._get_config(), config, dict(**kwargs))
        all_config = self._remove_meta(all_config)
        self.warnings = {}
        # add configuration items to the class as attributes
        self.__dict__.update(all_config)

        # do some API checks
        if webapp:
            try:
                all_config = merge_data(config, dict(**kwargs))
                self._api_checks(all_config)
            except AttributeError as e:
                self.warnings['no_api_check'] = warning_messages['no_api_check'] % (self.__class__.__name__, e)

    def _remove_meta(self, config):
        """
        The meta block in our json files is solely used for comments, data pedigree, and other metadata. As such,
        nothing in the engine should read it or need it; removing meta should do nothing to calculations but
        reduce the possibility of errors when looping through a dictionary (#3791).

        This routine (recursively) removes meta blocks from the configuration

        Parameters
        ----------
        config: dict
            A fully populated dictionary, ready to be added to the attributes of whatever class called this function

        Returns
        -------
        config: dict
            A fully populated dictionary, cleaned of meta blocks
        """
        if 'meta' in config:
            del config['meta']
        for item in config:
            # if the item itself is a dictionary, recurse inside for a meta block
            if isinstance(config[item], dict):
                config[item] = self._remove_meta(config[item])

        return config


    def as_dict(self):
        """
        Return dict representation of instance configuration. If self.api_parameters exists,
        use it to determine what to return.  otherwise, just use self.__dict__ to provide everything.

        Also scrub astropy quantities.
        """
        if hasattr(self, "api_parameters"):
            d = {}
            for p in self.api_parameters:
                d[p] = getattr(self, p)
        else:
            d = self.__dict__
        for item in d:
            if type(d[item]) is astropy.units.quantity.Quantity:
                d[item] = d[item].value
            if type(d[item]) in [astropy.units.core.PrefixUnit, astropy.units.core.Unit]:
                d[item] = d[item].to_string()
        return d

    def _get_config(self):
        """
        Read default configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # use this trick to key the configuration file name off the name of the instantiated subclass
        objname = self.__class__.__name__.lower()
        config_file = "defaults/%s_defaults.json" % objname
        config_path = pkg_resources.resource_filename(__name__, config_file)
        config = io.read_json(config_path, raise_except=True)
        return config

    def _api_checks(self, conf):
        """
        Check input configuration against self.api_parameters to make sure all the expected
        parameters are being set and no unrecognized parameters are being passed in.

        Parameters
        ----------
        conf: dict
            Engine input API format dict
        """
        # first, go through self.api_parameters and make sure they're all there
        for p in self.api_parameters:
            if p not in conf:
                msg = "API parameter, %s, missing from input for %s. Using default value of %s." % (
                    p,
                    self.__class__.__name__,
                    getattr(self, p)
                )
                self.warnings["%s_%s" % (self.__class__.__name__.lower(), p)] = msg

        all_pars = copy.deepcopy(self.api_parameters)
        if hasattr(self, "api_ignore"):
            all_pars.extend(self.api_ignore)

        # now go through the input and flag anything we don't know about
        for c in conf:
            if c not in all_pars and "ui_" not in c:
                msg = "Unsupported configuration parameter, %s, being passed to %s." % (c, self.__class__.__name__)
                self.warnings["%s_unsupported_%s" % (self.__class__.__name__.lower(), c)] = msg



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

from .telescope import Telescope
from .instrument import Instrument

class Roman(Telescope):

    """
    Currently a dummy class for directory/file discovery, but could eventually contain Roman-specific methods
    """
    pass


class RomanInstrument(Instrument):

    """
    Generic Roman Instrument class
    """

    def __init__(self, mode=None, config={}, **kwargs):
        telescope = Roman()
        # these are the required sections and need to be passed via API in webapp mode
        self.instrument_pars = {}
        self.instrument_pars['detector'] = ["nexp", "ngroup", "nint", "readout_pattern", "subarray"]
        self.instrument_pars['instrument'] = ["aperture", "disperser", "filter", "instrument", "mode"]
        self.api_parameters = list(self.instrument_pars.keys())

        # these are required for calculation, but ok to live with config file defaults
        self.api_ignore = ['dynamic_scene', 'max_scene_size', 'scene_size']

        Instrument.__init__(self, telescope=telescope, mode=mode, config=config, **kwargs)


class WFI(RomanInstrument):

    """
    Currently, the Roman WFI requires only one method beyond those provided by the generic Instrument class
    """
    def _loadpsfs(self):
        """
        Short-wavelength filters need PSFs that don't have the cold pupil mask (0.48-1.6 microns)
        Long-wavelength filters need PSFs that DO have the cold pupil mask (0.8-2.2 microns)
        Because the ranges overlap, we need to switch between PSF libraries.
        The mask is specifically baked into the filters, so selecting by filter seems appropriate
        """
        if self.instrument['mode'] == 'imaging':
            if self.instrument['filter'] in self.cold_pupil:
                psf_key = "%s%s" % (self.instrument['aperture'], 'lw')
            else:
                psf_key = "%s%s" % (self.instrument['aperture'], 'sw')
        elif self.instrument['mode'] == 'spectroscopy':
            psf_key = "%s%s" % (self.instrument['aperture'], 'lw')
        else:
            message = "Invalid mode specification: {}".format(str(self.instrument['mode']))
            raise EngineInputError(value=message)

        self.psf_library = self._load_psf_library(psf_key)


class IFU(RomanInstrument):

    """
    Currently the Roman IFU (deprecated) requires no extra methods beyond those provided by the generic Instrument class
    """
    pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

from .config import DefaultConfig
from .custom_exceptions import EngineInputError
from .scene import Scene
from .strategy import StrategyFactory


class Observation(DefaultConfig):

    """
    This class collects the pieces required to make an ETC simulated observation.
    """

    def __init__(self, scene=None, instrument=None, strategy=None, config={}, webapp=False, **kwargs):
        DefaultConfig.__init__(self, config=config, **kwargs)

        # Initialize the observation with the default source and the default strategy
        # relevant for the instrument/mode combo. The user can also pass an initializing source.
        if scene is None:
            self.scene = Scene(webapp=webapp)
        else:
            self.scene = scene

        # make sure we're given a properly configured Instrument
        if instrument is not None:
            self.instrument = instrument
            self.mode = self.instrument.mode
        else:
            msg = "Must provide configured Instrument subclass."
            raise EngineInputError(value=msg)

        # use strategy we're given or fall back to default for the given instrument
        if strategy is not None:
            self.strategy = strategy
        else:
            self.strategy = StrategyFactory(self.instrument, webapp=webapp)

    def get_random_seed(self):
        """
        Return the configured seed for the random number generator.
        """
        return self.random_seed



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

from .telescope import Telescope
from .instrument import Instrument
import astropy.units as u
import stsynphot as st
import numpy as np


class HST(Telescope):

    def get_ote_eff(self, wave):
        """
        Temporary function; allow stsynphot to handle all of the graph table functionality in one go.
        """
        return np.ones_like(wave)



class HSTInstrument(Instrument):

    """
    Generic HST Instrument class
    """
    def __init__(self, mode=None, config={}, **kwargs):
        telescope = HST()
        self.instrument_pars = {}
        self.instrument_pars['detector'] = ["nexp", "ngroup", "nint", "readout_pattern", "subarray"]
        self.instrument_pars['instrument'] = ["aperture", "disperser", "detector", "filter", "instrument", "mode"]
        self.api_parameters = list(self.instrument_pars.keys())

        # these are required for calculation, but ok to live with config file defaults
        self.api_ignore = ['dynamic_scene', 'max_scene_size', 'scene_size']
        Instrument.__init__(self, telescope=telescope, mode=mode, config=config, **kwargs)

    def _get_disperser(self):
        if self.instrument['disperser'] is None:
            return None
        elif "cenwave" in self.instrument:
            return "{},{},{},{}".format(self.instrument['instrument'], self.instrument['detector'], self.instrument['disperser'], self.instrument['cenwave'])
        else:
            return "{},{},{}".format(self.instrument['instrument'], self.instrument['detector'], self.instrument['disperser'])

    def _get_filter_key(self):
        if self.instrument['filter'] is None:
            return None
        else:
            return "{},{},{}".format(self.instrument['instrument'], self.instrument['detector'], self.instrument['filter'])


    def _get_cdbs(self, wave, key):
        """
        HST data is available in the $PYSYN_CBDS trees, so it should be loaded differently.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength array onto which the throughput curve is to be interpolated
        key: str
            Key to get filename from the $PYSYN_CDBS tree

        Returns
        -------
        eff: numpy.ndarray or float
            If ref file exists, return efficiency(wave), else return 1.0
        """
        if 'None' not in key:
            bp = st.spectrum.band(key)
            eff = bp(wave*u.micron).value #don't want to pass the Quantity
        else:
            # If it's explicitly None, there is no element; pass through 1.0
            eff = np.ones_like(wave)*1.0
        return eff

    def get_filter_eff(self, wave):
        """
        Temporary function; allow stsynphot to handle all of the graph table functionality in one go.
        """
        return np.ones_like(wave)

    def get_disperser_eff(self, wave):
        """
        Temporary function; allow stsynphot to handle all of the graph table functionality in one go.
        """
        return np.ones_like(wave)

    def get_detector_qe(self, wave):
        """
        Temporary function; allow stsynphot to handle all of the graph table functionality in one go.
        """
        return np.ones_like(wave)

    def get_internal_eff(self,wave):
        """
        This is the only temporary function that will actually do anything. It relies on
        stsynphot for graph table functionality
        
        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Disperser efficiency as a function of wave
        """
        obsmode = self._get_disperser() if self._get_filter_key() is None else self._get_filter_key()
        bp = st.spectrum.band(obsmode)
        eff = bp(wave*u.micron).value #don't want to pass the Quantity

        return eff


class COSFUV(HSTInstrument):

    """
    Special methods unique to HST COS/FUV
    """
    def _get_filter_key(self):
        if self.instrument['filter'] is None:
            return None
        else:
            return "cos,{},{}".format(self.instrument['detector'], self.instrument['filter'])

    def _get_disperser(self):
        if self.instrument['disperser'] is None:
            return None
        elif "cenwave" in self.instrument:
            return "cos,{},{},{}".format(self.instrument['detector'], self.instrument['disperser'], self.instrument['cenwave'])
        else:
            return "cos,{},{}".format(self.instrument['detector'], self.instrument['disperser'])


    def _loadpsfs(self):
        """
        The COS PSFs differ greatly by focus position, and we thus need to 
        know lifetime position (currently only LP4), disperser, and cenwave.
        """
        if self.instrument['mode'] == "spectroscopy":
            psf_key = "{}{}{}{}".format(self.instrument['aperture'], "lp4", self.instrument['disperser'], self.instrument['cenwave'])
        else:
            psf_key = self.instrument['aperture']
        self.psf_library = self._load_psf_library(psf_key)


class WFC3IR(HSTInstrument):

    """
    Special methods unique to HST WFC3IR
    """
    def _get_filter_key(self):
        if self.instrument['filter'] is None:
            return None
        else:
            return "wfc3,{},{}".format(self.instrument['detector'], self.instrument['filter'])


class WFC3UVIS(HSTInstrument):

    """
    Special methods unique to HST WFC3UVIS
    """
    def _get_filter_key(self):
        if self.instrument['filter'] is None:
            return None
        else:
            return "wfc3,{},{}".format(self.instrument['detector'], self.instrument['filter'])

class ACS(HSTInstrument):

    """
    Special methods unique to HST ACS
    """
    pass

class STIS(HSTInstrument):

    """
    Special methods unique to HST STIS
    """
        
    def get_wave_blaze(self):
        """
        Get wavelength vector used in the grating efficiency (blaze) file

        Returns
        -------
        wave_blaze: numpy.ndarray
            Wavelength vector from the grating efficiency file
        """
        key = self._get_disperser()
        if key is not None:
            bp = st.spectrum.band(key)
        return bp.waveset.to_value(u.micron)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import numpy as np
import numpy.ma as ma
from .custom_exceptions import EngineInputError
from .constants import MIN_CLIP


class ExposureSpecification:

    """
    Encapsulates all data associated with exposure specification
    """

    def __init__(self, pattern, ngroup, nint, nexp, tframe, nframe=1, nsample=1, nsample_skip=0, tfffr=0,
                 subarray='Full', nskip=0, nprerej=0, npostrej=0, nreset1=1, ndrop1=0, ndrop3=0, nreset2=1,
                 frame0=False, det_type='h2rg'):
        """
        Create a generic Exposure Specification.

        Inputs
        ------
        pattern: str
            name of readout pattern

        ngroup: int
            number of groups per integration

        nint: int
            number of integrations per exposure

        nexp: int
            number of exposures

        tframe: float
            number of seconds per frame

        nframe: int (optional, default 1)
            number of frames per group

        nsample: int (optional, default 1)
            number of samples averaged while reading a single pixel.
            Used by MIR detectors.

        nsample_skip: int (optional, default 0)
            number of samples skipped while reading a single pixel.
            Used by MIR detectors.

        tfffr: float
            Extra time factor

        subarray: str (optional, default 'Full')
            name of subarray to be read.
            If None, readout is full frame

        nskip: int (optional, default 0)
            number of skipped frames per group. Only
            supported by some readout patterns.

        nprerej: int (optional, default 0)
            The MIR detectors (i.e. MIRI) may reject the first and/or last groups of a ramp.

        npostrej: int (optional, default 0)
            Number of rejected frames at the end of the integration

        nreset1: int (optional, default 1)
            Number of reset frames before the integration starts

        ndrop1: int (optional, default 0)
            Number of frames dropped at the beginning of an integration

        ndrop3: int (optional, default 0)
            Number of frames dropped at the end of an integration

        nreset2: int (optional, default 1)
            Number of reset frames after the integration ends

        frame0: bool (optional, default False)
            Some instruments can downlink the frame0 read. Set to True if frame0 is used in the ramp fit.

        det_type: string
            Detector type, used to determine timing calculation types.

        """

        # Eventually we might do some value checking at this point
        # to ensure that the specified values are consistent with the
        # named pattern. For now, pattern is unused.

        # Required parameters
        self.det_type = det_type
        self.pattern = pattern
        self.ngroup = ngroup
        self.nint = nint
        self.nexp = nexp
        self.tframe = tframe
        self.tfffr = tfffr

        # Optional parameters
        self.nframe = nframe
        self.subarray = subarray
        self.nskip = nskip
        self.nprerej = nprerej
        self.npostrej = npostrej
        self.nreset1 = nreset1
        self.nreset2 = nreset2
        self.ndrop1 = ndrop1
        self.ndrop3 = ndrop3
        self.nsample = nsample
        self.nsample_skip = nsample_skip
        self.frame0 = frame0

        if det_type == 'sias':
            self.get_times_mir()
            # This is where we adjust values so we can still use the same
            # MULTIACCUM formula as for the NIR detectors. We need the effective
            # "average time per sample" for MIRI.
            self.tsample = self.tframe / (self.nsample + self.nsample_skip)
            # 'nsample_total' for MIRI is the total number of non-skipped samples X number of averaged frames.
            # Note that in practice, it currently never happens that both the number of samples and number of
            # averaged frames are >1 (i.e., no SLOWGRPAVG exists). However, this will deal with that situation,
            # should it occur.
            self.nsample_total = self.nframe * self.nsample

        elif det_type == 'h2rg':
            self.get_times_nir()
        else:
            raise ValueError('Unknown detector type %s' % (det_type))

        # Derived quantities
        self.nramps = self.nint * self.nexp

    def get_times_nir(self):
        """
        The time formulae are defined in Valenti et al. 2017, JWST-STScI-006013

        The equations in this method are duplicated in the front-end
        (workbook.js, update_detector_time_labels function). Please ensure
        that changes to these equations are reflected there.
        """
        self.tgroup = self.tframe * (self.nframe + self.nskip)

        # Equation 4
        self.measurement_time = self.tframe * self.nint * ((self.ngroup - 1) * (self.nframe + self.nskip))
        # Equation 5
        if self.frame0:
            self.measurement_time += 0.5 * self.nint * self.tframe * self.nframe

        # Equation 1
        self.exposure_time = (self.tfffr * self.nint) + self.tframe * (self.nreset1 + (self.nint - 1) * self.nreset2 +
                                self.nint * (self.ndrop1 + (self.ngroup - 1) * (self.nframe + self.nskip) +
                                self.nframe + self.ndrop3))
        # Equation 7
        self.saturation_time = self.tfffr + self.tframe * (self.ndrop1 +
                               (self.ngroup - 1) * (self.nframe + self.nskip) + self.nframe)

        self.duty_cycle = self.measurement_time / self.exposure_time
        self.total_exposure_time = self.nexp * self.exposure_time
        self.total_integrations = self.nexp * self.nint

    def get_times_mir(self):
        """
        The time formulae are defined in Valenti et al. 2017, JWST-STScI-006013. Note that we have to subtract the
        groups that are rejected (by the pipeline) from the measurement time (nprerej+npostrej). The saturation time
        conservatively considers the ramp saturated even if saturation occurs in a rejected frame.

        The equations in this method are duplicated in the front-end
        (workbook.js, update_detector_time_labels function). Please ensure
        that changes to these equations are reflected there.
        """

        self.tgroup = self.tframe * (self.nframe + self.nskip)

        # Valenti et al. 2017 defines the very bright regime for MIRI as ngroup>=5.
        # This could eventually be pulled out to data, rather than hardcoded.
        # Equation 6
        if self.ngroup >= 5:
            self.measurement_time = self.nint * (self.ngroup - 1 - (self.nprerej + self.npostrej)) * self.tgroup
        else:
            # very bright regime
            # Equation 10
            self.measurement_time = self.nint * (self.tfffr + self.ngroup * self.tgroup) \
                - self.nint * self.tframe * 0.5 * (self.nframe - 1)

        # Equation 3
        self.exposure_time = self.nint * (self.tfffr + self.ngroup * self.nframe * self.tframe)
        # Equation 8
        self.saturation_time = self.tfffr + self.ngroup * self.nframe * self.tframe

        self.duty_cycle = self.measurement_time / self.exposure_time
        self.total_exposure_time = self.nexp * self.exposure_time

        self.total_integrations = self.nexp * self.nint

    def get_unsaturated_groups(self, slope, fullwell, full_saturation=2):
        """
        Calculate the number of unsaturated groups in each pixel, given a specified slope and full well value.
        There is a minimum sensible value of unsaturated number of groups that defines full saturation.
        The default is 2, but this can potentially be set to a higher value, or even 1(!) for certain observing modes.
        It is not expected that a value of 0 can ever be anything but saturated.

        Parameters
        ----------
        slope: ndarray
            The measured slope of the ramp (ie, the rate) per pixel

        fullwell: positive integer
            The number of electrons defining a full well (beyond which the pixel reads become unusable for science
            due to saturation).

        full_saturation: positive integer
            The minimum number of groups allowed to define an unsaturated measurement.

        Returns
        -------
        unsat_ngroups: MaskedArray
            The number of unsaturated groups present in the ramp. The mask separates pixels that have full saturation
            from those that do not. The former will not have a valid noise measurement and cannot be used in a strategy.

        """

        # get the number of groups to saturation
        ngroups_to_sat = self.get_groups_before_sat(slope, fullwell)

        # the number of unsaturated groups is clipped by the ngroups_to_sat value for every pixel
        max_ngroups = ngroups_to_sat.clip(0,self.ngroup)

        # Mask values that are fully saturated - if the maximum before saturation is less than full_saturation, it's
        # clearly saturated.
        unsat_ngroups = ma.masked_less(max_ngroups, full_saturation)

        return unsat_ngroups

    def get_groups_before_sat(self, slope, fullwell):
        """
        This method returns a map of the maximum number of groups each pixel can be exposed to, before saturation.
        The formula is calculated by equating the time-to-saturation to the saturation time formula and isolating
        ngroup. The resulting fractional value is then rounded down to an integer.

        These equations are derived by setting saturation_time = time_to_saturation in equation 7 (for NIR H2RG
        detectors) or equation 8 (for MIR SiAs detectors) of the timing document, Valenti et al. 2017,
        JWST-STScI-006013, and solving for n_groups.

        Technically speaking, we are getting the groups BEFORE saturation by rounding the number of groups down to the
        nearest integer. If groups_before_sat is EXACTLY an integer (which is unlikely), we will get one partially
        saturated pixel.

        Parameters
        ----------
        slope: ndarray
            The measured slope of the ramp (ie, the rate) per pixel

        fullwell: positive integer
            The number of electrons defining a full well (beyond which the pixel reads become unusable for science
            due to saturation).
        """
        # we clip to 1e-10 to avoid division by zero errors even if the slope is tiny.
        time_to_saturation = fullwell / slope.clip(MIN_CLIP, np.max([MIN_CLIP, np.max(slope)]))
        if self.det_type == 'sias':
            groups_before_sat = (time_to_saturation - self.tfffr) / (self.nframe * self.tframe)
        elif self.det_type == 'h2rg':
            groups_before_sat = (((time_to_saturation - self.tfffr) / self.tframe) - self.nframe) / (self.nframe +
                                                                                                     self.nskip) + 1.
        else:
            raise ValueError('Unknown detector type %s' % (self.det_type))
        # The groups_before_sat value should never be less than 0 - that would imply saturation before we started
        # collecting any photons.
        groups_before_sat = np.maximum(groups_before_sat,np.zeros_like(groups_before_sat))

        return np.floor(groups_before_sat)

    def get_saturation_fraction(self, slope, fullwell):
        """
        This method returns a map of the fraction of saturation that each pixel reaches, given the current
        exposure setup.

        Parameters
        ----------
        slope: ndarray
            The measured slope of the ramp (ie, the rate) per pixel

        """
        # we clip to 1e-10 to avoid division by zero errors even if the slope is tiny.
        return self.saturation_time / (fullwell / slope.clip(MIN_CLIP, np.max([MIN_CLIP, np.max(slope)])))

    def slope_variance(self, rate, dark_current, readnoise, unsat_ngroups,
                       rn_fudge=1.0, excessp1=0.0, excessp2=0.0):
        """
        Calculate the variance of a specified MULTIACCUM slope.

        Inputs
        ------
        rate: ndarray
            The measured slope of the ramp (ie, the rate) per pixel
            This can be a one-dimensional wavelength spectrum (for a 1D ETC)
            or a three-dimensional cube ([wave,x,y] for a 3D ETC).

        dark_current: float
            Dark current (electrons/s).

        readnoise: float
            Readnoise per pixel.

        unsat_ngroups: ndarray
            Number of unsaturated groups for each pixel

        rn_fudge: float
            Fudge factor to apply to readnoise to match IDT results

        excessp1: float
             empirical correction for excess variance for long ramps

        excessp2: float
             empirical correction for excess variance for long ramps

        Returns
        -------
        slope_var: ndarray
            Variance associated with the input slope.
        slope_rn_var: ndarray
            The associated variance of the readnoise only
        slope_dc_var: ndarray
            The associated variance of the dark current only
        """

        # Rename variables for ease of comparison with Robberto (35).
        # The noise calculation depends on the pixel rate BEFORE IPC convolution. fp_pix_variance takes that pre-IPC
        # rate and scales it by the quantum yield and Fano factor to get the per-pixel variance in the electron rate.
        variance_per_pix = rate['fp_pix_variance']

        rn = readnoise

        # we discard any saturated groups. Copy unsat_ngroups, because we may modify it for rejected groups before using
        # it for noise calculations.
        n = ma.copy(unsat_ngroups)

        # removing with any pre- and post-rejected groups here. This is not stictly directly related to saturation, bu
        # behaves in a similar way (if there is <2 groups available, we cannot define a slope). The very bright regime
        # in which groups are not rejected can be implemented (in parts) by setting nprerej=npostrej=0.
        # Important: frames are not rejected in the Very Bright Regime (self.ngroup<5).
        # Therefore we disable it here in that regime.
        if ((self.nprerej != 0) or (self.npostrej != 0)) and (self.ngroup>=5):
            # If a group is already rejected because of saturation, there is no need to reject another at the end of
            # the ramp.
            unsat_pixels = (self.ngroup-np.ceil(n))<self.npostrej
            n[unsat_pixels] -= self.nprerej
            # we do have to reject any post-rejected frames from all pixels
            n -= self.npostrej

        # maybe nsample_sample_total and tsample are defined. If so then use them for the effect frame time and number
        # of reads (Issue #2996)
        # This is similar to two other places in this file
        if hasattr(self, 'nsample_total') and hasattr(self, 'tsample'):
            m = self.nsample_total
            tframe = self.tsample
        else:
            m = self.nframe # This does not include skipped frames!
            tframe = self.tframe

        tgroup = self.tgroup

        # Compute the variance of a MULTIACCUM slope using Robberto's formula (35). The rn_variance is also
        # calculated using the unsaturated number of groups.
        slope_rn_var = self.rn_variance(rn, unsat_ngroups=n, rn_fudge=rn_fudge, excessp1=excessp1, excessp2=excessp2)
        # The final slope variance slope may also be worse than the theoretical best value.
        # (also from Glasse et al. 2015, PASP 127 686).
        slope_var = (6. / 5.) * (n ** 2. + 1.) / (n * (n ** 2. - 1.)) * \
            ((variance_per_pix + dark_current) / tgroup) * \
            (1. - (5. / 3.) * (m ** 2. - 1.) / (m * (n ** 2. + 1.)) * (tframe / tgroup)) + \
            slope_rn_var
        # just the variance due to the dark current (remove poisson and readnoise)
        slope_dc_var = (6. / 5.) * (n ** 2. + 1.) / (n * (n ** 2. - 1.)) * \
            ((np.zeros_like(variance_per_pix) + dark_current) / tgroup) * \
            (1. - (5. / 3.) * (m ** 2. - 1.) / (m * (n ** 2. + 1.)) * (tframe / tgroup))

        # The default fill value for masked arrays is a finite number, so convert to ndarrays, and fill with NaNs
        # to make sure missing values are interpreted as truly undefined downstream.

        slope_var = ma.filled(slope_var, fill_value=np.nan)
        slope_rn_var = ma.filled(slope_rn_var, fill_value=np.nan)
        slope_dc_var = ma.filled(slope_dc_var, fill_value=np.nan)

        return slope_var, slope_rn_var, slope_dc_var

    def rn_variance(self, readnoise, unsat_ngroups=None, rn_fudge=1.0, excessp1=0.0, excessp2=0.0):
        """
        Calculate the variance due to read noise only.

        Inputs
        ------
        readnoise: float
          Readnoise per pixel.

        unsat_ngroups: ndarray
           The number of unsaturated groups.

           If unsat_ngroups not supplied, the approximation is that the read noise is
           negligible for pixels with signal rates high enough to saturate in part of the ramp.
           This is probably always a good approximation.

        readnoise: float
          Readnoise per pixel.

        rn_fudge: float
            fudge factor for readnoise on the slope as it may be worse than the theoretical best value

        excessp1: float
             empirical correction for excess variance for long ramps

        excessp2: float
             empirical correction for excess variance for long ramps

        Returns
        -------
        var_rn: ndarray if unsat_ngroups is supplied, float if unsat_ngroups is not supplied.
           Variance associated with the read noise only.

        """
        if unsat_ngroups is None:
            n = self.ngroup - (self.nprerej + self.npostrej)
        else:
            n = unsat_ngroups

        # It was decided that only one of the fudge factors are allowed to be used, not both.
        # Issue #2261#issuecomment-260737726
        if rn_fudge != 1.0 and (excessp1 != 0.0 or excessp2 != 0.0):
            raise ValueError('Only one of rn_fudge or excessp1/p2 maybe used, not both.')

        rn = readnoise

        # maybe nsample_sample_total is defined. If so then use them for the effect frame time and number of reads
        # Issue #2996
        # This is similar to two other places in this file
        if hasattr(self, 'nsample_total'):
            m = self.nsample_total
        else:
            m = self.nframe # This does not include skipped frames!

        tgroup = self.tgroup

        var_rn = 12. * rn ** 2. / (m * n * (n ** 2. - 1.) * tgroup ** 2.)

        # The readnoise on the slope may be worse than the theoretical best value
        # (see Glasse et al. 2015, PASP 127 686).
        if rn_fudge != 1:
            var_rn *= rn_fudge

        # Include the empirical correction for excess variance for long ramps
        # (Issue #2091)
        if excessp1 != 0.0 or excessp2 != 0.0:
            excess_variance = (12.0 * (n-1)/(n+1) * excessp1**2 - excessp2/np.sqrt(m)) / ((1 - n) * tgroup)**2
            var_rn += excess_variance
        return var_rn


class ExposureSpecificationTA(ExposureSpecification):

    """
    Target acquisition uses a special exposure specification and noise formula. Readout patterns have the same
    model as a regular MULTIACCUM exposure.

    Pontoppi 2017 Technical Report, SOCCER number tbd
    """

    def __init__(self, pattern, ngroup, ngroup_extract, nint, nexp, tframe, nframe=1, nsample=1, tfffr=0,
                 nsample_skip=1, subarray='Full', nskip=0, nprerej=0, npostrej=0, nreset1=1, ndrop1=0, ndrop3=0,
                 nreset2=1, det_type='h2rg'):
        """
        Create a Target Acquisition Exposure Specification.

        Inputs
        ------
        pattern: str
            name of readout pattern

        ngroup: int
            number of groups per integration

        ngroup_extrac: int
            number of groups actually used by the TA script (it ignores most of them). 3 is a
            common value. Note that this has to be <ngroup.

        nint: int
            number of integrations per exposure

        nexp: int
            number of exposures

        tframe: float
            number of seconds per frame

        nframe: int (optional, default 1)
            number of frames per group

        nsample: int (optional, default 1)
            number of samples averaged while reading a single pixel.
            Used by MIR detectors.

        nsample_skip: int (optional, default 0)
            number of samples skipped while reading a single pixel.
            Used by MIR detectors.

        subarray: str (optional, default 'Full')
            name of subarray to be read.
            If None, readout is full frame

        nskip: int (optional, default 0)
            number of skipped frames per group. Only
            supported by some readout patterns.

        nprerej: int (optional, default 0)
            The MIR detectors (i.e. MIRI) reject the first and last reads of a ramp (and maybe more eventually).
            For checking saturation, we need to handle them separately.  We need to account for signal accrued
            during frames rejected before the ramp, but don't care about the signal afterwards.

        npostrej: int (optional, default 0)
            Frames rejected after a ramp don't count towards saturation time, but need to be included in total exposure time.

        frame0: bool (optional, default False)
            Some instruments can downlink the frame0 read. Set to True if frame0 is used in the ramp fit.

        det_type: string (optional, default 'h2rg')
            Detector type, timing calculation is dependent on detector type

        """

        ExposureSpecification.__init__(self, pattern, ngroup, nint, nexp, tframe, nframe=nframe,
                 subarray=subarray, nskip=nskip, nsample_skip=nsample_skip, tfffr=tfffr,
                 nprerej=nprerej, npostrej=npostrej, nreset1=nreset1, ndrop1=ndrop1, ndrop3=ndrop3,
                 nreset2=nreset2, det_type=det_type)

        # Target acqs only use a subset of the groups
        self.ngroup_extract = ngroup_extract

    def slope_variance(self, rate, dark_current, readnoise, unsat_ngroups,
                       rn_fudge=1.0, excessp1=0.0, excessp2=0.0):
        """
        Calculate the slope variance

        Inputs
        ------
        rate: ndarray
            Rate

        dark_current: float


        readnoise: float
          Readnoise per pixel.

        unsat_ngroups: ndarray
           The number of unsaturated groups.

           If unsat_ngroups not supplied, the approximation is that the read noise is
           negligible for pixels with signal rates high enough to saturate in part of the ramp.
           This is probably always a good approximation.

        rn_fudge: float
            fudge factor for readnoise on the slope as it may be worse than the theoretical best value

        excessp1: float
             empirical correction for excess variance for long ramps

        excessp2: float
             empirical correction for excess variance for long ramps

        Returns
        -------
        var_rn: ndarray if unsat_ngroups is supplied, float if unsat_ngroups is not supplied.
           Variance associated with the read noise only.

        """

        # Rename variables for ease of comparison with Robberto (35).
        # The noise calculation depends on the pixel rate BEFORE IPC convolution. fp_pix_variance takes that pre-IPC
        # rate and scales it by the quantum yield and Fano factor to get the per-pixel variance in the electron rate.
        variance_per_pix = rate['fp_pix_variance']

        rn = readnoise
        n = unsat_ngroups  # we discard any saturated groups
        # maybe nsample_sample_total and tsample are defined. If so then use them for the effect frame time and number
        # of reads (Issue #2996)
        # This is similar to two other places in this file
        if hasattr(self, 'nsample_total') and hasattr(self, 'tsample'):
            m = self.nsample_total
            tframe = self.tsample
        else:
            m = self.nframe # This does not include skipped frames!
            tframe = self.tframe

        nextract = self.ngroup_extract

        # Throw error if the number of extract groups is >= number of groups.
        if self.ngroup_extract > self.ngroup:
            raise EngineInputError("Number of groups to extract, {}, must be less than or equal to the number of groups {}".format(
                self.ngroup_extract, self.ngroup
            ))

        tgroup = self.tgroup

        # The effective group time. This is the time from the first group to the next extracted group. For instance,
        # if ngroups=11 and nextract=3, textract = tgroup * 5
        textract = tgroup * (n-1)/(nextract-1)

        # The TA script will take the minimum of co-subtracted pairs. This has non-Gaussian error properties,
        # but a Monte Carlo Simulation indicates an improvement of 2/3 on the variance for the minimum of two
        # correlated pairs in a single ramp (as compared to a factor 1/2 for the average). It would be good to
        # demonstrate this analytically, but that has not yet been accomplished.

        # Number of possible pairs.
        npairs = nextract-1
        if npairs == 1:
            min_factor = 1.0
        elif npairs == 2:
            min_factor = 2./3.
        else:
            raise EngineInputError("Target acquisition currently only supports 2 or 3 extracted groups.")

        # Modification of MULTIACCUM formula for n=2 and a revised "effective" group time that takes into
        # account that only some groups obtained may be used to create the TA images.
        slope_rn_var = min_factor * (2*rn**2)/m/textract**2
        slope_var = min_factor * (variance_per_pix+dark_current)/textract * (1.-(1./3.)*(m**2-1)/m *
                                                                             tframe/textract) + slope_rn_var

        # And for dark current alone:
        slope_dc_var = min_factor * (np.zeros_like(variance_per_pix)+dark_current)/textract * (1.-(1./3.)*(m**2-1)/m *
                                                                             tframe/textract)

        # The default fill value for masked arrays is a finite number, so convert to ndarrays, and fill with NaNs
        # to make sure missing values are interpreted as truly undefined downstream.

        slope_var = ma.filled(slope_var, fill_value=np.nan)
        slope_rn_var = ma.filled(slope_rn_var, fill_value=np.nan)

        return slope_var, slope_rn_var, slope_dc_var



#!/usr/bin/env python

# setuptools is required
from setuptools import setup

import subprocess

with open('README.md') as fp:
    description = fp.read()

# Versioning
try:
    DEVBUILD = subprocess.check_output(["git", "describe", "--tags"])
    with open('pandeia/engine/helpers/DEVBUILD', 'wb') as out:
        out.write(DEVBUILD)
except (subprocess.CalledProcessError) as err:
    print(err)

setup(
    # The package
    name="pandeia.engine",
    version="1.6",
    packages=["pandeia",
              "pandeia.engine",
              "pandeia.engine.defaults",
              "pandeia.engine.helpers",
              "pandeia.engine.helpers.schema"],
    # For PyPI
    description='Pandeia 3D Exposure Time Calculator compute engine',
    long_description=description,
    author='Adric Riedel, Klaus Pontoppidan, Craig Jones, Christopher Sontag, Oi In Tam Litten, Tim Pickering',
    #author_email='jwsthelp.stsci.edu',
    url='https://jwst.etc.stsci.edu',
    classifiers=["Intended Audience :: Science/Research",
                 "License :: OSI Approved :: BSD License",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.6",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.8",
                 "Topic :: Scientific/Engineering :: Astronomy",
                 "Topic :: Software Development :: Libraries :: Python Modules"],
    # Other notes
    package_data={
        "pandeia.engine.defaults": ["*.json"],
        "pandeia.engine.helpers": ["DEVBUILD"]
        },
    include_package_data=True,
    install_requires=[
        "numpy>=1.17",
        "scipy",
        "astropy>=4",
        "photutils",
        "synphot",
        "stsynphot",
        "six",
        "setuptools"
    ],
    zip_safe=False
)
